#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <unistd.h>
#include <iostream>
#include <vector>
#include <queue>
#include <utility>
#include <algorithm>
using namespace std;

#include "common.h"
#include "marshal.h"
#include "update.h"

// skratka. pouzivajte v zistiTah().
#define vykonaj(p) vykonajPrikaz(mapa, stav, mojTah, (p))
#define par(a,b) make_pair(a, b)

#define FARBA 8

vector<vector<int> > farba;
vector<vector<vector<int> > > dist;
int A[8]={1,1,1,0,-1,-1,-1,0},
    B[8]={1,0,-1,-1,-1,0,1,1};

int ja;   // ktory hrac som
Mapa mapa;
Stav stav;
Tah mojTah;

int bfs(int x, int y, int c) {
  bool apa[mapa.h][mapa.w];
  for (int i=0; i<mapa.h; i++) {
    for (int j=0; j<mapa.w; j++) {
      apa[i][j]=false;
    }
  }
  apa[y][x]=true;
  queue<pair<pair<int,int>,int> > QQ;
  QQ.push(par(par(x,y),0) );
  while (!QQ.empty()) {
    int X=QQ.front().first.first,
	Y=QQ.front().first.second,
	z=QQ.front().second;
    if (farba[Y][X]==c) {
      return z;
    }
    for (int i=0; i<8; i++) {
      if (mapa.priechodne(X+A[i],Y+B[i])) {
	if (!apa[Y+B[i]][X+A[i]]) {
	  apa[Y+B[i]][X+A[i]]=true;
	  QQ.push(par(par(X+A[i],Y+B[i]),z+1) );
	}
      }
    }
    QQ.pop();
  }
  return -1; //z neznameho dovodu to nevyslo
}


double skore(int x, int y) {
  double max=0;
  FOREACH (q, stav.questy) {
    double l=0;
    FOREACH (p, stav.predmety) {
      if (q->ziadanyPredmet != p->typ) continue;
      /////////////////////////////////////////
      double vyh = (q->zlato)/(1 + dist[y][x][farba[p->y][p->x] ] + dist[q->cielY][q->cielY][farba[p->y][p->x] ]);
      if (vyh>l) l=vyh; 
    }
    if (l>max) max=l;
  }
  return max;
}


// main() zavola tuto funkciu, ked nacita mapu
void inicializuj() {
  // (sem patri vas kod)
  
  //urob minimapu
  farba.resize(mapa.h);
  dist.resize(mapa.h);
  for (int i=0; i<mapa.h; i++) {
    farba[i].resize(mapa.w);
    dist[i].resize(mapa.w);
  }
  for (int i=0; i<mapa.h; i++) {
    for (int j=0; j<mapa.w; j++) {
      farba[i][j]=0;
    }
  }
  int k=1;
  queue<pair<int,int> > QQ;
  for (int i=FARBA; i<mapa.h; i+=2*FARBA+1) {
    for (int j=FARBA; j<mapa.w; j+=2*FARBA+1) {
      if (mapa.priechodne(j,i)) {
	QQ.push(par(j,i));
	farba[i][j]=k;
	k++;
      }
    }
  }
  for (int i=0; i<mapa.h; i++) {
    for (int j=0; j<mapa.w;j++) {
      dist[i][j].resize(k);
    }
  }
  while (!QQ.empty()) {
    int x=QQ.front().first,
	y=QQ.front().second;
    for (int i=0; i<8; i++) {
      if (mapa.priechodne(x+A[i],y+B[i])) {
	if (farba[y+B[i]][x+A[i]]==0) {
	  farba[y+B[i]][x+A[i]]=farba[y][x];
	  QQ.push(par(x+A[i],y+B[i]));
	}
      }
    }
    QQ.pop();
  } //minimapa hotova!
  
  for (int i=0; i<mapa.h; i++) {
    for (int j=0; j<mapa.w; j++) {
      if (mapa.priechodne(j,i)) {
	for (int l=1; l<k; l++) {
	  dist[i][j][l]=bfs(j,i,l);
	}
      }
    }
  }
}


// main() zavola tuto funkciu, ked chce vediet, ake prikazy chceme vykonat
void zistiTah() {
  // (sem patri vas kod)
  // priklad: if (vykonaj(Prikaz::chod(niekto, x, y))) { /* podarilo sa */ }

  // verbuj hrdinu
  int X, Y;
  do {
    X=rand()%mapa.w;
    Y=rand()%mapa.h;
  } while (!vykonaj(Prikaz::verbuj(0,X,Y)) ) ;

  
  bool chcemTah; //chcem Ta
  do {
    chcemTah=false;
    FOREACH (it, stav.zivocichy) {
      if (it->ktorehoHraca != ja) continue;
      /////////////////////////////////////
      int smer=-1;
      double vyhodnost=skore(it->x, it->y);
      for (int i=0; i<8; i++) {
	double S=skore(it->x + A[i], it->y + B[i]);
	if (S>vyhodnost) {
	  smer=i; vyhodnost=S;
	}
      }
      if (vykonaj(Prikaz::chod(it->id,it->x+A[smer],it->y+B[smer])) ) chcemTah=true;
    }
  } while (chcemTah) ;
}


int main() {
  // v tejto funkcii su vseobecne veci, nemusite ju menit (ale mozte).

  unsigned int seed = time(NULL) * getpid();
  srand(seed);

  nacitaj(cin, ja);
  nacitaj(cin, mapa);
  fprintf(stderr, "START ja=%d, pid=%d, seed=%u\n", ja, getpid(), seed);
  inicializuj();

  while (cin.good()) {
    nacitaj(cin, stav);
    mojTah = zaciatokTahu(stav, ja);
    zistiTah();
    uloz(cout, mojTah.prikazy);
    cout << ".\n" << flush;   // bodka a flush = koniec odpovede
  }

  return 0;
}

