#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <unistd.h>
#include <iostream>
#include <vector>
#include <queue>
#include <string>
#include <sstream>
#include <iomanip>

using namespace std;

#include "common.h"
#include "marshal.h"
#include "update.h"


int ja;   // ktory hrac som
Mapa mapa;
vector<vector<int> > pole;

Stav stav;
Tah mojTah;

map <string, bool> debug_flags;
string curr_debug_flag = "Main";

#define sstream stringstream

sstream ss;

void debug (string message, string flag = ""){
    if (flag == "") flag = curr_debug_flag;

    if (debug_flags.count(flag) == 0 ) {
        cerr << "[" + flag + "]: " + message << endl;
        return;
    }

    if (!debug_flags[flag]) return;

    cerr << "[" + flag + "]: " + message << endl;
}
void debug (){
    string flag = curr_debug_flag;
    string temp = ss.str();
    ss.str("");

    if (debug_flags.count(flag) == 0 ) {
        cerr << "[" + flag + "]: " + temp << endl;
        return;
    }

    if (!debug_flags[flag]) return;

    cerr << "[" + flag + "]: " + temp << endl;
}

void debug_init (){
    debug_flags["Main"] = true;
    debug_flags["Initialize"] = true;
    debug_flags["Actualize"] = true;
    debug_flags["Target select"] = true;
    debug_flags["Clock"] = true;
    debug_flags["Path finding"] = true;
    debug_flags["BFS 1"] = false;
    debug_flags["Make beware"] = true;
}

void set_debug_flag (string flag){
    curr_debug_flag = flag;
}

#define setdf set_debug_flag


// skratka. pouzivajte v zistiTah().
#define vykonaj(p) vykonajPrikaz(mapa, stav, mojTah, (p))

struct Bod {
    int x, y;

    template <class T>
    T& get (vector<vector<T>> &data);
    int get (Mapa data);

    Bod operator+ (Bod a) {
        return Bod(this->x+a.x, this->y+a.y);
    }

    bool operator== (Bod a) {
        return (a.x == this->x) && (a.y == this->y);
    }

    Bod (int x, int y) {
        this->x = x;
        this->y = y;
    }

    template <class T>
    Bod (T data) {
        this->x = data.x;
        this->y = data.y;
    }

    Bod () {}    //Shut up Code::Blocks
};

template <class T>
T& Bod::get (vector<vector<T>> &data) {
    return data[x][y];
}

int Bod::get (Mapa data) {
    return data.zisti(x, y);
}



vector <Bod> mesta;
vector <int> mojiHrdinovia;
Quest targetQuest;
vector<Bod> d_cka= {{-1,0},{-1,1},{-1,-1},{0,1},{0,-1},{1,0},{1,1},{1,-1}};


int sgn(int c);
int countHeroes (int hrac=-1);
void inicializujZoznamMiest ();
void aktualizujMojichHrdinov();
void bfs(vector<vector<int> > &nazov, Bod bod);
void bfs(vector<vector<int> > &nazov, Bod bod, vector <vector<int> > &beware);
int zisti(Bod);
Zivocich findLifeByID(int ID);

void inicializuj() {
    debug_init();

    setdf("Initialize");

    debug("started");
    inicializujZoznamMiest();
    targetQuest.zlato = -100000;
    targetQuest.ziadanyPredmet= -11;
    debug ("finished");
}


bool verbujem = false;

void zistiTah() {
    setdf("Clock");
    ss << stav.cas;
    debug();

    setdf ("Actualize");

    debug ("started");
    aktualizujMojichHrdinov();
    debug("finished");

    setdf ("Make beware");
    debug("started");
    vector <vector <int> > beware (mapa.w, vector<int> (mapa.h, 0));

    for (auto monster : stav.zivocichy) {
        if (monster.ktorehoHraca == ja) continue;

        Bod prisera (monster);
        prisera.get(beware) = 1;
    }

    for (auto item: stav.predmety){
        Bod(item).get(beware) = 1;
    }


/*    for (auto &vec : beware){
        for (auto i: vec){
            ss << i << " ";
        }
        debug();
    }*/

    debug("finished");

    setdf("Hero select");

    for (auto ponuka: stav.pracovnePonuky){
        if (ponuka.ktorehoHraca != ja) continue;
        bool collision = false;
        for (auto life : stav.zivocichy){
            if (Bod(life) == Bod(ponuka)) collision = true;
        }

        if (! collision) verbujem = true;
    }


    if ((countHeroes() == 0) && (verbujem == false)) {
        int r = rand() % mesta.size();
        if (vykonaj(Prikaz::verbuj(HRDINA_LUMP, mesta[r].x, mesta[r].y))) {
            debug("Verbujem hrdinu");
        }
    }


    if (countHeroes() == 0) return;


    Zivocich hrdina = findLifeByID(mojiHrdinovia[0]);
    ss << "Id: " << hrdina.id << " hraca: " << hrdina.ktorehoHraca << " pos: " << hrdina.x << " " << hrdina.y << " neseny predmet: " << hrdina.nesenyPredmet;
    debug();

    setdf ("Target select");
    debug ("started");

    bool found = false;
    debug ("Hladam quest");

    for (auto quest: stav.questy) {
        bool identic = true;

        if (targetQuest.cielX != quest.cielX) identic = false;
        if (targetQuest.cielY != quest.cielY) identic = false;
        if (targetQuest.ziadanyPredmet != quest.ziadanyPredmet) identic = false;

        found |= identic;
    }

    if (found) {
        ss << "Quest is still existing: " << targetQuest.cielX << " "<< targetQuest.cielY << " " << targetQuest.ziadanyPredmet;
        debug ();
    }


    for (auto pred: stav.predmety){
        ss << pred.typ << " ";
    }
    debug();


    Bod me (hrdina.x, hrdina.y);
    Bod target (0, 0);

    debug ("starting bfs");
    vector<vector<int> > bfsdata;
    bfs (bfsdata, me);
    debug ("ending bfs");


    if (! found) {
        debug("Error 404: Quest not found! Is it finished ?");
        if (stav.questy.size() == 0) return;
        targetQuest.zlato = -100000;

        for (auto quest: stav.questy) {
            bool existItem = false;
            Bod itemLoc;
            for (auto item: stav.predmety){
                if (quest.ziadanyPredmet == item.typ) {
                    itemLoc= Bod(item);

                    existItem = true;
                }
            }

            if (!existItem) continue;
            debug ("valid quest");

            if (targetQuest.zlato < (quest.zlato- itemLoc.get(bfsdata)) ) {
                targetQuest = quest;
                found = true;
            }
        }
        if (!found) return;
    }





    if ((hrdina.nesenyPredmet != -1) && (hrdina.nesenyPredmet != targetQuest.ziadanyPredmet)) {
        Bod temp = Bod(hrdina) + d_cka[rand()% 8];

        vykonaj(Prikaz::poloz(mojiHrdinovia[0], temp.x, temp.y));
        debug ("kladiem predmet");
        hrdina = findLifeByID(mojiHrdinovia[0]);
    }

    if (hrdina.nesenyPredmet == targetQuest.ziadanyPredmet) {

        debug ("vraciam sa do mesta");
        target = Bod (targetQuest.cielX, targetQuest.cielY);

        bool askar = false;
        if (!(Bod(hrdina) == target)) {
            while(vykonaj(Prikaz::utoc(mojiHrdinovia[0], targetQuest.cielX, targetQuest.cielY))) {
                debug("Askar uhni");
                askar = true;
            }
        }
        if (askar) debug ("Askar uhnil");

        bool closeToCity = false;
        for (auto smer: d_cka){
            Bod kand = smer + target;
            if (kand == Bod(hrdina)){
                closeToCity = true;
            }
        }

        bool stupidItem = false;
        for (auto item: stav.predmety){
            if (Bod(item) == target) stupidItem = true;
        }
        /*
        if (closeToCity && stupidItem) {
            for (auto smer: d_cka){
                vykonaj(Prikaz::poloz(mojiHrdinovia[0], ))

            }

        }*/

        vykonaj(Prikaz::poloz(mojiHrdinovia[0], targetQuest.cielX, targetQuest.cielY));
    } else {
        debug("hladam item");
        for (auto predmet: stav.predmety) {
            if (predmet.typ != targetQuest.ziadanyPredmet) continue;

            Bod newtar (predmet.x, predmet.y);
            if (target.get(bfsdata) > newtar.get(bfsdata)) target = newtar;
        }
    }

    ss << "target " << target.x << " " << target.y;
    debug();
    ss << "tar param: "<< target.get(mapa) << " " << target.get(beware);
    debug();


    debug("finished");

    setdf("Path finding");
    debug("started");

    ss << "Beware: ";
    for (auto x: d_cka){
        Bod temp = x+Bod(hrdina);
        ss << temp.get(beware) << " ";
    }
    debug();

    bfs (bfsdata, target, beware);

    /*for (auto vec : bfsdata){
        for (auto i: vec){
            ss << i << " ";
        }
        debug();
    }*/

    bool poholSomSa;
    do {
        debug ("new turn");
        me = Bod (hrdina);

        poholSomSa = false;

        Bod dir = Bod(0, 0);
        int Min = 100000;



        for(auto smer: d_cka){
            Bod temp = me + smer;
            ss << temp.get(bfsdata) << " ";

            if (temp.get(bfsdata) < Min){
                dir = temp;
                Min = temp.get(bfsdata);
            }

        }
        debug();
        //debug();

        ss << "Direction: " << dir.x << " " << dir.y;
        debug();

        if (Bod (0, 0) == dir) continue;
        if (vykonaj(Prikaz::chod(hrdina.id, dir.x, dir.y))) poholSomSa = true;
        hrdina = findLifeByID(mojiHrdinovia[0]);
    } while (poholSomSa);
}



void bfs(vector<vector<int> > &nazov, Bod bod) {

    nazov.clear();
    nazov.resize(mapa.w,vector<int> (mapa.h,100908));
    queue<Bod> haldicka;


    haldicka.push(bod);
    bod.get(nazov)=0;
    int pocet = 0;

    while(!haldicka.empty()) {
        Bod pom = haldicka.front();
        haldicka.pop();

        for(auto pos: d_cka) {
            Bod tar = pos + pom;
            /*if((tar.get(mapa) > 0) && (tar.get(nazov) > pom.get(nazov) + 1) ) {
                tar.get(nazov) = pom.get(nazov)+1;
                haldicka.push(tar);
            }*/

            if((mapa.zisti(tar.x, tar.y) > 0) && (nazov[tar.x][tar.y] > nazov[pom.x][pom.y] + 1) ) {
                nazov[tar.x][tar.y] = nazov[pom.x][pom.y]+1;
                haldicka.push(tar);
            }
        }
    }
}

void bfs(vector<vector<int> > &nazov, Bod bod, vector<vector<int> > &beware) {

    nazov.clear();
    nazov.resize(mapa.w,vector<int> (mapa.h,100908));
    queue<Bod> haldicka;


    haldicka.push(bod);
    bod.get(nazov)=0;
    int pocet = 0;

    while(!haldicka.empty()) {
        Bod pom = haldicka.front();
        haldicka.pop();

        for(auto pos: d_cka) {
            Bod tar = pos + pom;
            /*if((tar.get(mapa) > 0) && (tar.get(nazov) > pom.get(nazov) + 1) ) {
                tar.get(nazov) = pom.get(nazov)+1;
                haldicka.push(tar);
            }*/

            if((mapa.zisti(tar.x, tar.y) > 0) && (nazov[tar.x][tar.y] > nazov[pom.x][pom.y] + 1) ) {
                if (beware[tar.x][tar.y] == 1) continue;
                nazov[tar.x][tar.y] = nazov[pom.x][pom.y]+1;
                haldicka.push(tar);
            }
        }
    }
}

int sgn(int c) {
    if (c < 0) return -1;
    if (c > 0) return 1;
    return 0;
}

void aktualizujMojichHrdinov() {
    for (int i=0; i<stav.zivocichy.size(); i++) {
        if (stav.zivocichy[i].ktorehoHraca == ja) {
            mojiHrdinovia.push_back(stav.zivocichy[i].id);
        }
    }
}

int zisti(Bod a) {
    return mapa.zisti(a.x,a.y);
}

int countHeroes (int hrac) {
    if (hrac == -1) return mojiHrdinovia.size();

    int mojiHrdinovia = 0;
    for (auto it: stav.zivocichy) {
        if (it.ktorehoHraca == ja) mojiHrdinovia++;
    }

    return mojiHrdinovia;
}


void inicializujZoznamMiest () {
    for (int y = 0; y < mapa.h; y++) {
        for (int x = 0; x < mapa.w; x++) {
            if (mapa.zisti(x, y) == MAPA_MESTO) {
                mesta.push_back(Bod(x, y));
            }
        }
    }
}
Zivocich findLifeByID(int ID){
    for (auto ziv: stav.zivocichy){
        if (ziv.id == ID) return ziv;
    }

    return stav.zivocichy[0];
}

int main() {
    // v tejto funkcii su vseobecne veci, nemusite ju menit (ale mozte).

    unsigned int seed = time(NULL) * getpid();
    srand(seed);

    nacitaj(cin, ja);
    nacitaj(cin, mapa);
    fprintf(stderr, "START ja=%d, pid=%d, seed=%u\n", ja, getpid(), seed);
    inicializuj();

    while (cin.good()) {
        nacitaj(cin, stav);
        mojTah = zaciatokTahu(stav, ja);
        zistiTah();
        uloz(cout, mojTah.prikazy);
        cout << ".\n" << flush;   // bodka a flush = koniec odpovede
    }

    return 0;
}


