#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <unistd.h>
#include <iostream>
#include <vector>
#include <queue>
#include <set>
using namespace std;

#include "common.h"
#include "marshal.h"
#include "update.h"


int ja;   // ktory hrac som
Mapa mapa;
Stav stav;
Tah mojTah;

typedef pair<int, int> pii;


// skratka. pouzivajte v zistiTah().
#define vykonaj(p) vykonajPrikaz(mapa, stav, mojTah, (p))

vector<pair<int, int> > mesta;


// main() zavola tuto funkciu, ked nacita mapu
void inicializuj() {
  // (sem patri vas kod)
  for (int y = 0; y < mapa.h; y++) {
    for (int x = 0; x < mapa.w; x++) {
      if (mapa.zisti(x, y) == MAPA_MESTO) {
        fprintf(stderr, "mam mesto %d %d\n", x, y);
        mesta.push_back(make_pair(x, y));
      }
    }
  }
  fprintf(stderr, "init end\n");
}


int sgn(int c) {
  if (c < 0) return -1;
  if (c > 0) return 1;
  return 0;
}

vector<vector<int> > GetDist(int sx, int sy, set<pii>& banned) {
  vector<vector<int> > dd(mapa.w, vector<int>(mapa.h, -1));

  queue<pair<int, int> > fr;
  fr.push(make_pair(sx, sy));
  dd[sx][sy] = 0;
  while (!fr.empty()) {
    pii x = fr.front(); fr.pop();
    for (int i = 0; i < 8; i++) {
      int nx = x.first + DX[i];
      int ny = x.second + DY[i];
      if (mapa.priechodne(nx, ny) && dd[nx][ny] == -1 && 
          banned.count(make_pair(nx, ny)) == 0) {
        fr.push(make_pair(nx, ny));
        dd[nx][ny] = dd[x.first][x.second]+1;
      }
    }
  }
  return dd;
}

vector<vector<int> > GetDistArea(vector<pair<int, int> > s, set<pii>& banned) {
  vector<vector<int> > dd(mapa.w, vector<int>(mapa.h, -1));

  queue<pair<int, int> > fr;
  for (int i = 0; i < s.size(); i++) {
    if (banned.count(s[i])) continue;
    if (!mapa.priechodne(s[i].first, s[i].second)) continue;
    fr.push(s[i]);
    dd[s[i].first][s[i].second] = 0;
  }
  while (!fr.empty()) {
    pii x = fr.front(); fr.pop();
    for (int i = 0; i < 8; i++) {
      int nx = x.first + DX[i];
      int ny = x.second + DY[i];
      if (mapa.priechodne(nx, ny) && dd[nx][ny] == -1 && 
          banned.count(make_pair(nx, ny)) == 0) {
        fr.push(make_pair(nx, ny));
        dd[nx][ny] = dd[x.first][x.second]+1;
      }
    }
  }
  return dd;}

vector<vector<int> > GetDist2(int sx, int sy, set<pii>& banned) {
  vector<vector<int> > dd(mapa.w, vector<int>(mapa.h, -1));

  queue<pair<int, int> > fr;
  fr.push(make_pair(sx, sy));
  dd[sx][sy] = 0;
  while (!fr.empty()) {
    pii x = fr.front(); fr.pop();
    for (int i = 0; i < 8; i++) {
      int nx = x.first + DX[i];
      int ny = x.second + DY[i];
      if (dd[nx][ny] == -1 && mapa.priechodne(nx, ny) &&
          banned.count(make_pair(nx, ny)) == 0) {
        fr.push(make_pair(nx, ny));
        dd[nx][ny] = dd[x.first][x.second]+1;
      }
      if (dd[nx][ny] == -1) {
        dd[nx][ny] = dd[x.first][x.second]+1;
      }
    }
  }
  return dd;
}


void Move(int targetx, int targety, int id, int sx, int sy, set<pii>& banned) {
  vector<vector<int> > dd = GetDist(targetx, targety, banned);
  int mc = 0;
  while (mc < 8 && (sx != targetx || sy != targety)) {
    mc++;
    for (int i = 0; i < 8; i++) {
      int nx = sx + DX[i];
      int ny = sy + DY[i];
      if (mapa.priechodne(nx, ny) && dd[nx][ny] == dd[sx][sy] - 1) {
        if (!vykonaj(Prikaz::chod(id, nx, ny))) {
          if (!vykonaj(Prikaz::utoc(id, nx, ny))) {
            fprintf(stderr, "nieco sa posralo %d %d\n", nx, ny);
          } else {
            while (vykonaj(Prikaz::utoc(id, nx, ny))) ;
          }
          return;
        } else {
          sx = nx;
          sy = ny;
        }
      }
    }
  }
}

void MoveAndDrop(int targetx, int targety, int id, int sx, int sy,
                 set<pii>& banned) {
  vector<pair<int, int> > targets;
  set<pii> target_set;
  targets.push_back(make_pair(targetx, targety));
  target_set.insert(make_pair(targetx, targety));
  for (int i = 0; i < 8; i++) {
    targets.push_back(make_pair(targetx+DX[i], targety+DY[i]));
    target_set.insert(make_pair(targetx+DX[i], targety+DY[i]));
  }
  if (target_set.count(make_pair(sx, sy))) {
    for (int i = 0; i < 8; i++) {
      if (vykonaj(Prikaz::poloz(id, targetx + DX[i], targety + DY[i])))
        return;
    }
  }
  fprintf(stderr, "drop dist start\n");
  vector<vector<int> > dd = GetDistArea(targets, banned);
  fprintf(stderr, "drop dist end\n");
  int mc = 0;
  while (mc < 8 && (sx != targetx || sy != targety)) {
    mc++;
    for (int i = 0; i < 8; i++) {
      int nx = sx + DX[i];
      int ny = sy + DY[i];
      if (mapa.priechodne(nx, ny) && dd[nx][ny] == dd[sx][sy] - 1) {
        if (target_set.count(make_pair(nx, ny))) {
          if (!vykonaj(Prikaz::poloz(id, nx, ny))) {
            fprintf(stderr, "nieco sa posralo pri drope %d %d %d\n", id, nx, ny);

          } else {
            fprintf(stderr, "dobry drop %d %d %d\n", id, nx, ny);
          }
        
          return;
        }
        if (!vykonaj(Prikaz::chod(id, nx, ny))) {
          fprintf(stderr, "nieco sa posralo %d %d\n", nx, ny);
          return;
        } else {
          sx = nx;
          sy = ny;
        }
      }
    }
  }
}

bool first_move = true;
map<int, pair<int, int> > targets;
map<int, int> kill_targets;
map<int, pair<int, int> > drop_targets;

// main() zavola tuto funkciu, ked chce vediet, ake prikazy chceme vykonat
void zistiTah() {
  // id zivocicha - pozicia
  fprintf(stderr, "startjem tah\n");
  set<pii> empty_banned;
  map<int, vector<int> > predmety;
  for (int i = 0; i < stav.predmety.size(); i++) {
    predmety[stav.predmety[i].typ].push_back(i);
  }
/*  map<pair<int, int>, vector<vector<int> > > dist_mats;
  for (int i = 0; i < stav.questy.size(); i++) {
    int tx = stav.questy[i].cielX;
    int ty = stav.questy[i].cielY;
    if (dist_mats.count(make_pair(tx, ty)) == 0) {
      dist_mats[make_pair(tx, ty)] = GetDist(tx, ty, empty_banned);
    }
    vector<vector<int> > &dd = dist_mats[make_pair(tx, ty)];
    for (int j = 0; j < predmety[stav.questy[i].ziadanyPredmet].size(); j++) {
      int sx = stav.predmety[predmety[stav.questy[i].ziadanyPredmet][j]].x;
      int sy = stav.predmety[predmety[stav.questy[i].ziadanyPredmet][j]].y;
      fprintf(stderr, "dist quest %d\n", dd[sx][sy]);
    }
  }*/
  
  vector<Zivocich> moji;
  set<pair<int, int> > banned, banned_plus_items;
  for (int i = 0; i < stav.zivocichy.size(); i++) {
    if (stav.zivocichy[i].ktorehoHraca == ja) {
      fprintf(stderr, "moj hp %d/%d\n", stav.zivocichy[i].hp,
          kHrdinaHp[2]);
      moji.push_back(stav.zivocichy[i]);
    }
    else
      banned.insert(make_pair(stav.zivocichy[i].x, stav.zivocichy[i].y));
  }
  banned_plus_items = banned;
  for (int i = 0; i < stav.predmety.size(); i++)
    banned_plus_items.insert(make_pair(stav.predmety[i].x, stav.predmety[i].y));
  fprintf(stderr, "mam %d\n", moji.size());
  map<pair<int, int>, vector<vector<int> > > dist_mat;

  if (moji.size() > 0) {
    fprintf(stderr, "moja pozicia %d %d nesiem %d\n", moji[0].x, moji[0].y,
            moji[0].nesenyPredmet);
    if (moji[0].nesenyPredmet == -1) {
      if (drop_targets.count(moji[0].id) != 0)
          drop_targets.erase(moji[0].id);
      int tx = moji[0].x;
      int ty = moji[0].y;
      
      // zistime vzdialenost od agenta ku vsetkemu moznemu

      if (dist_mat.count(make_pair(tx, ty)) == 0)
        dist_mat[make_pair(tx, ty)] = GetDist(tx, ty, empty_banned);
      vector<vector<int> >& dd = dist_mat[make_pair(tx, ty)];
      double maxval = -1000000;
      pii mindist_p;
      int mindist_pris = -1;
      
      // pre kazdy quest hladam itemy medzi predmetmi a priserami
      for (int i = 0; i < stav.questy.size(); i++) {
        int mx = stav.questy[i].cielX;
        int my = stav.questy[i].cielY;
        if (dist_mat.count(make_pair(mx, my)) == 0)
          dist_mat[make_pair(mx, my)] = GetDist(mx, my, empty_banned);
        vector<vector<int> >& md = dist_mat[make_pair(mx, my)];
        for (int j = 0; j < predmety[stav.questy[i].ziadanyPredmet].size(); j++) {
          int sx = stav.predmety[predmety[stav.questy[i].ziadanyPredmet][j]].x;
          int sy = stav.predmety[predmety[stav.questy[i].ziadanyPredmet][j]].y;
          fprintf(stderr, "dist quest panak %d %d\n", dd[sx][sy], 
                  stav.questy[i].zlato);
          int dist = dd[sx][sy] + md[sx][sy];
          double val = (double)stav.questy[i].zlato / dist;
          if (val > maxval) {
            maxval = val;
            mindist_p = make_pair(sx, sy);
            mindist_pris = -1;
          }
        }
        for (int j = 0; j < stav.zivocichy.size(); j++) {
          if (stav.zivocichy[j].ktorehoHraca == ja) continue;
          if (stav.zivocichy[j].nesenyPredmet == stav.questy[i].ziadanyPredmet) {
            fprintf(stderr, "dist quest panak-prisera %d %d hp %d predmet %d\n",
                dd[stav.zivocichy[j].x][stav.zivocichy[j].y],
                stav.questy[i].zlato, stav.zivocichy[j].hp,
                stav.questy[i].ziadanyPredmet);
            double dist = dd[stav.zivocichy[j].x][stav.zivocichy[j].y];
            // TODO: toto poriadne
            bool prezijem = true;
            if (prezijem) {
              dist += 2*stav.zivocichy[j].hp*kApUtoc[stav.zivocichy[j].obrana[2]][0];
              dist += md[stav.zivocichy[j].x][stav.zivocichy[j].y];
              if (stav.zivocichy[j].hp > moji[0].hp)
                dist *= 2; 
               if (stav.zivocichy[j].ktorehoHraca != -1)
                dist *= 2;
              double val = (double)stav.questy[i].zlato / dist;
              if (val > maxval) {
                maxval = val;
                mindist_pris = stav.zivocichy[j].id;
                mindist_p = make_pair(stav.zivocichy[j].x, stav.zivocichy[j].y);
              }
            }
          }
        }
      }
      if (maxval > 0) {
        if (mindist_pris != -1) {
          if (targets.count(moji[0].id) == 0 || rand() % 3 == 0 
              || kill_targets[moji[0].id] == mindist_pris || true) {
            targets[moji[0].id] = mindist_p;
            kill_targets[moji[0].id] = mindist_pris;
          }
        } else {
          if (targets.count(moji[0].id) == 0 || rand() % 3 == 0 || true)
            targets[moji[0].id] = mindist_p;
        }
        fprintf(stderr, "target %d: %d %d\n", moji[0].id,
            targets[moji[0].id].first, targets[moji[0].id].second);
      }
    } else {
      if (targets.count(moji[0].id))
        targets.erase(moji[0].id);
      if (kill_targets.count(moji[0].id))
        kill_targets.erase(moji[0].id);
      if (drop_targets.count(moji[0].id) != 0)
          drop_targets.erase(moji[0].id);
      double best_drop = -10000000;
      
      int tx = moji[0].x;
      int ty = moji[0].y;
      
      if (dist_mat.count(make_pair(tx, ty)) == 0)
        dist_mat[make_pair(tx, ty)] = GetDist(tx, ty, empty_banned);
      vector<vector<int> >& dd = dist_mat[make_pair(tx, ty)];
      for (int i = 0; i < stav.questy.size(); i++) {
        if (stav.questy[i].ziadanyPredmet == moji[0].nesenyPredmet) {
          int dist = dd[stav.questy[i].cielX][stav.questy[i].cielY];
          double val = (double)stav.questy[i].zlato / dist;
          if (val > best_drop) {
            best_drop = val;
            drop_targets[moji[0].id] = make_pair(
                stav.questy[i].cielX, stav.questy[i].cielY);
            fprintf(stderr, "drop_target %d: %d %d\n", moji[0].id,
                    drop_targets[moji[0].id].first, drop_targets[moji[0].id].second);
          }
        }
      }
      if (drop_targets.count(moji[0].id) == 0) {
        fprintf(stderr, "useless predmet pokladam\n");
        for (int i = 0; i < 8; i++) {
          if (vykonaj(Prikaz::poloz(
                      moji[0].id, moji[0].x + DX[i], moji[0].y + DX[i])))
            break;
        }
      }
    }
  } else {
    fprintf(stderr, "verbujem\n");
    int rm = rand()%mesta.size();
    vykonaj(Prikaz::verbuj(2, mesta[rm].first, mesta[rm].second));
    //vykonaj(Prikaz::verbuj(2, mesta[0].first, mesta[0].second));
  }

  for (int i = 0; i < moji.size(); i++) {
    if (targets.count(moji[i].id)) {
      Move(targets[moji[i].id].first, targets[moji[i].id].second,
           moji[i].id, moji[i].x, moji[i].y, banned);
    }
    if (drop_targets.count(moji[i].id)) {
      MoveAndDrop(drop_targets[moji[i].id].first, drop_targets[moji[i].id].second,
                  moji[i].id, moji[i].x, moji[i].y, banned_plus_items);
      drop_targets.erase(moji[0].id);
    }
  }

  fprintf(stderr, "este mi zostalo %d Action Points\n", mojTah.ap);
}


int main() {
  // v tejto funkcii su vseobecne veci, nemusite ju menit (ale mozte).

  unsigned int seed = time(NULL) * getpid();
  srand(seed);

  nacitaj(cin, ja);
  nacitaj(cin, mapa);
  fprintf(stderr, "START ja=%d, pid=%d, seed=%u\n", ja, getpid(), seed);
  inicializuj();

  while (cin.good()) {
    nacitaj(cin, stav);
    mojTah = zaciatokTahu(stav, ja);
    zistiTah();
    uloz(cout, mojTah.prikazy);
    cout << ".\n" << flush;   // bodka a flush = koniec odpovede
  }

  return 0;
}

