
Čo je proboj
------------

Proboj je počítačová hra, ktorej hráčmi nie ste vy, ale programy, čo napíšete.

Tentoraz je téma proboja **RPG** (čítaj **hra na hrdinov**). Hráči môžu ovládať
hrdinov, biť sa s potvorami (alebo medzi sebou), hľadať poklady a plniť questy
za tučné finančné odmeny.


Zdrojáky
--------

Štandardný hráč čiže klient (v adresári `klienti/template`) sa skladá z jediného
zdrojáku `main.cpp`. Ale môžte ho rozdeliť aj na viacero. Dokonca môžte použiť
aj iný jazyk (aj keď to je trochu problematické, v prípade záujmu sa ma
opýtajte).

V serveri je tiež zopár zdrojákov, čo vás bude zaujímať.

- `common.h` obsahuje základné štruktúry, čo váš klient dostane k dispozícii.
- `update.cpp` a `update.h` obsahujú všetky herné konštanty, a tiež
  implementáciu väčšiny herných pravidiel, takže ak v pravidlách nie je niečo
  jasné, skúste sa tam pozrieť.
- v `main.cpp` sú tiež nejaké pravidlá (ako sa ťahá apod.), ale to je asi menej
  dôležité.

Kľudne si prečítajte aj ostatné zdrojáky, ja sa len poteším, ale pri kódení
vášho klienta vám asi nepomôžu.


Ako kódiť klienta
-----------------

Skopírujte obsah `klienti/template` do iného adresára a niečo v ňom nakóďte.

V koreni proboju spustite `make`, čím všetko skompilujete. (Ak váš klient nie je
vnútri `klienti`, nastavte v jeho `Makefile` správny `SERVERDIR` a spustite
`make` aj v ňom.)

Potom spustite `./server/server zaznamy/01 mapy/mapa1.ppm klienti/vasklient
klienti/vasklient klienti/hlupy` To spustí hru s troma hráčmi (vaším, druhým
vaším a hlúpym) a uloží záznam do `zaznamy/01`. Ten si môžete pozrieť s príkazom
`./observer/observer zaznamy/01`.

Server sa vášho klienta pýta, čo chce robiť. Ak klient neodpovie včas, bude
automaticky zabitý a reštartovaný. Prvýkrát dostane viac času, aby sa mohol
inicializovať.

Keď server spustíte u vás, je to len na skúšku. Na hlavnom počítači to beží na
ostro. Je tam aj webové rozhranie, cez ktoré môžete uploadovať vašich klientov.
Uploadujú sa zdrojáky a tie sa potom skompilujú (konkrétne sa spustí `make
naserveri SERVERDIR=/adresar/kde/je/server`).


Ako sa ťahá
-----------

Server najprv klientom určí náhodné poradie ťahov. Vždy, keď sa potom budú
klienti číslovať, použije sa toto poradie, čiže "klient 0" je ten, čo prvý
ťahá, nie nutne ten, ktorého spravila prvá družinka.

Potom sa hráči v danom poradí do konca hry striedajú. Vždy, keď váš hráč
príde na ťah, dostane niekoľko AP čiže Action Points (alebo Akčné Puntíky) a
za ne môže dávať svojím hrdinom príkazy. Každý príkaz stojí niekoľko AP. Keď
sa AP minú, hráč musí počkať, kým na ďalší ťah nedostane nové.

Takže dokopy to vyzerá nejako takto:

* Server vám povie momentálny `struct Stav`, čiže kompletnú informáciu o tom,
  ako práve vyzerá situácia.
* Na vašom klientovi vznikne `struct Tah`, v ktorom je, koľko máte AP, a zoznam
  vašich príkazov (zatiaľ prázdny).
* Váš klient môže vykonávať príkazy. Ak sa príkaz nepodarí (či už lebo máte
  málo AP, alebo je to úplná blbosť), nič sa nestane. Ak sa podarí, prejaví sa
  to na klientskom `Stav`e a `Tah`u. (To sa stane len na klientovi, server
  o tých príkazoch ešte nevie.)
* Keď je klient spokojný a skončí svoj ťah, pošle serveru zoznam príkazov, čo
  si vybral. (A nič iné, takže nemôže len tak meniť ozajstný `Stav`.)
* Server si rovnako spraví `Tah` a rovnako vykoná príkazy, čo mu prišli. Ak sa
  nejaký nepodarí, berie to ako pokus o sabotáž a daného klienta pre istotu
  reštartuje. (Vykonávanie príkazov nezávisí od náhody, takže na serveri by sa
  malo stať to isté, ako na klientovi.)
* Na záver server všelijaké pozadiové veci -- pohne potvorami, vygeneruje nové
  questy a podobne. (Toto už od náhody závisieť môže.)

Jeden z dôsledkov je, že na klientovi môžete robiť "undo". `Stav` a `Tah` si
odložíte do pomocných premenných, a keď ich neskôr z tých premenných vytiahnete,
všetko bude, ako predtým. (Serveru je to jedno, tomu posielate zoznam príkazov
až na konci.)


Pravidlá hry
------------

**Mapa**: je štvorčeková a dá sa na nej hýbať do 8 smerov. Sú na nej mestá.
A aj iné veci, ale mestá sú najdôležitejšie. Po mape sa prechádzajú hrdinovia
a potvory (spolu živočíchy). A niekedy tam sú pohodené všelijaké predmety.

**Hrdinovia**: každý hráč ich môže mať najviac štyroch. Nemajú samostatné AP --
ak chce jeden spraviť viac príkazov, ostatní spravia menej. Dajú sa verbovať v
krčme (čiže v každom meste), ale od začiatku verbovania niečo trvá, kým sa tam
nejaký hrdina objaví. Hrdinovia majú rôzne typy. Typ určuje, akou zbraňou útočí,
a ako sa proti rôznym zbraniam vie brániť. Z každého typu môže hráč mať najviac
jedného. Hrdinovia nemajú level ani nič podobné.

**Jedlo**: hráči majú v každom ťahu k dispozícii 25 AP. Na začiatku ťahu musí
za každého hrdinu zaplatiť 1 AP na jedlo. Za zvyšné AP môže robiť príkazy.

**Questy**: sú buď tvaru "prenes predmet z mesta do mesta", "poraz potvoru a
dones do mesta jej poklad", alebo "nájdi na mape položený poklad a dones".
V každom prípade na splnenie questu musí hrdina zohnať požadovaný predmet a
doniesť ho do daného mesta. Ak ten predmet nájde iný hrdina, môže ho skúsiť
poraziť a predmet mu zobrať.

**Príkazy**: hrdina môže buď prejsť na vedľajšie políčko (a pritom zobrať
predmet na danom políčku), alebo zaútočiť na vedľajšie políčko, alebo položiť
svoj predmet na vedľajšie políčko. (Na jednom políčku nikdy nie je aj živočích,
aj predmet.) Čím viac nejakého príkazu hrdina za kolo spraví, tým je to drahšie.
(Ak sa nejaký hrdina pohne dvakrát za kolo, je to drahšie, ako dvaja raz.)

**Konštanty**: všetky online služby zvyknú varovať: "Neverte nám, ak budeme
tvrdiť, že potrebujeme vaše heslo!" Podobne varujem: "Nepýtajte sa ma na žiadne
konštanty, a aj keby som vám nejaké povedal, neverte im!" To sa týka aj všetkých
čísel vyššie. Konštanty sa môžu a budú meniť. Pravdu nájdete iba
v `server/update.cpp`. Nedávajte do svojho klienta konkrétne čísla, ale volajte
ich menom (tie začínajúce na `k`).
