#include "marshal.h"
#include "common.h"
#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_image.h"
#include "observer.h"
#include <set>
#include <vector>

SDL_Color PLAYER_COLORS[MAX_PLAYERS] = {
  (SDL_Color){0xFF, 0x00, 0x00, 0xFF},
  (SDL_Color){0x00, 0xFF, 0x00, 0xFF},
  (SDL_Color){0x00, 0x00, 0xFF, 0xFF},
  (SDL_Color){0xFF, 0xFF, 0x00, 0xFF},
  (SDL_Color){0x00, 0xFF, 0xFF, 0xFF},
  (SDL_Color){0xFF, 0x00, 0xFF, 0xFF},
  (SDL_Color){0xFF, 0xFF, 0xFF, 0xFF},
  (SDL_Color){0x00, 0x00, 0x00, 0xFF},
  (SDL_Color){0x80, 0x40, 0x00, 0xFF},
  (SDL_Color){0x00, 0x40, 0x80, 0xFF},
};

SDL_Surface* player_figures[MAX_PLAYERS];
SDL_Surface* player_figures_reverse[MAX_PLAYERS];
SDL_Surface* player_small_figures[MAX_PLAYERS];
SDL_Surface* monster_figures;
SDL_Surface* monster_figures_reverse;
SDL_Surface* monster_small_figures;

SDL_Surface* item_small_icon;
SDL_Surface* item_carried_icon;
SDL_Surface* item_big_icon;

SDL_Surface* blood_icon;

int getUnitTileIndex(int type) {
  switch (type) {
    case HRDINA_BOJOVNIK: return 0;
    case HRDINA_ZALESAK: return 1;
    case HRDINA_LUMP: return 2;
    case HRDINA_CARODEJ: return 3;
    case HRDINA_KLERIK: return 4;
  }
  
  return 0;
}

std::set<int> who_moved_right;

void loadFigures() {
  /* nacitaj obrazky */
  for (int k = 0; k < NUM_PLAYERS; k++) {
    player_figures[k] = IMG_Load(getFile("figures/units.png"));
    player_figures_reverse[k] = IMG_Load(getFile("figures/units.png"));
    player_small_figures[k] = IMG_Load(getFile("figures/unit_icons.png"));
  }
  
  monster_figures = IMG_Load(getFile("figures/monsters.png"));
  monster_figures_reverse = IMG_Load(getFile("figures/monsters.png"));
  monster_small_figures = IMG_Load(getFile("figures/monster_icons.png"));
  
  item_small_icon = IMG_Load(getFile("figures/item_small.png"));
  item_carried_icon = IMG_Load(getFile("figures/item_carried.png"));
  item_big_icon = IMG_Load(getFile("figures/item_big.png"));
  
  blood_icon = IMG_Load(getFile("figures/blood.png"));
  
  /* prirad farby figurkam */
  for (int k = 0; k < NUM_PLAYERS; k++) {
    SDL_LockSurface(player_figures[k]);
    
    for (int i = 0; i < player_figures[k] -> w; i++)
      for (int j = 0; j < player_figures[k] -> h; j++) {
        Uint32* pixel = ((Uint32*)(player_figures[k] -> pixels) + (j * player_figures[k] -> w) + i);
        Uint8 r, g, b;
        
        SDL_GetRGB(*pixel, player_figures[k] -> format, &r, &g, &b);
        
        if (r != 0xFF || g != 0xFF || b != 0xFF) {
          r = (Uint8)(r + ((int)PLAYER_COLORS[PLAYER_ORDER[k]].r - r) / 3);
          g = (Uint8)(g + ((int)PLAYER_COLORS[PLAYER_ORDER[k]].g - g) / 3);
          b = (Uint8)(b + ((int)PLAYER_COLORS[PLAYER_ORDER[k]].b - b) / 3);
        
          *pixel = SDL_MapRGB(player_figures[k] -> format, r, g, b);
        }
      }
    
    SDL_UnlockSurface(player_figures[k]);    
  }
  
  /* prirad farby malym ikonkam */
  for (int k = 0; k < NUM_PLAYERS; k++) {
    SDL_LockSurface(player_small_figures[k]);
    
    for (int i = 0; i < player_small_figures[k] -> w; i++)
      for (int j = 0; j < player_small_figures[k] -> h; j++) {
        Uint32* pixel = ((Uint32*)(player_small_figures[k] -> pixels) + (j * player_small_figures[k] -> w) + i);
        Uint8 r, g, b;
        
        SDL_GetRGB(*pixel, player_small_figures[k] -> format, &r, &g, &b);
        
        if (r != 0xFF || g != 0xFF || b != 0xFF) {
          r = (Uint8)(r + ((int)PLAYER_COLORS[PLAYER_ORDER[k]].r - r) / 1);
          g = (Uint8)(g + ((int)PLAYER_COLORS[PLAYER_ORDER[k]].g - g) / 1);
          b = (Uint8)(b + ((int)PLAYER_COLORS[PLAYER_ORDER[k]].b - b) / 1);
        
          *pixel = SDL_MapRGB(player_small_figures[k] -> format, r, g, b);
        }
      }
    
    SDL_UnlockSurface(player_small_figures[k]);
  }

  /* vytvor zrkadlove obrazy */
  
  SDL_LockSurface(monster_figures);
  SDL_LockSurface(monster_figures_reverse);
    
  for (int m = 0; m < NUM_MONSTERS; m++)
    for (int i = 0; i < ZOOM_TILE_WIDTH; i++)
      for (int j = 0; j < ZOOM_TILE_HEIGHT; j++) {
        Uint32* pixel = ((Uint32*)(monster_figures -> pixels) + (j * monster_figures -> w) + m * ZOOM_TILE_WIDTH + i);
        Uint32* target = ((Uint32*)(monster_figures_reverse -> pixels) + (j * monster_figures_reverse -> w) + (m + 1) * ZOOM_TILE_WIDTH - i - 1);
        
        *target = *pixel;
      }

  SDL_UnlockSurface(monster_figures_reverse);      
  SDL_UnlockSurface(monster_figures);
  
  for (int k = 0; k < NUM_PLAYERS; k++) {
    SDL_LockSurface(player_figures[k]);      
    SDL_LockSurface(player_figures_reverse[k]);      
    
    for (int m = 0; m < NUM_UNITS; m++)
      for (int i = 0; i < ZOOM_TILE_WIDTH; i++)
        for (int j = 0; j < ZOOM_TILE_HEIGHT; j++) {
          Uint32* pixel = ((Uint32*)(player_figures[k] -> pixels) + (j * player_figures[k] -> w) + m * ZOOM_TILE_WIDTH + i);
          Uint32* target = ((Uint32*)(player_figures_reverse[k] -> pixels) + (j * player_figures_reverse[k] -> w) + (m + 1) * ZOOM_TILE_WIDTH - i - 1);
          
          *target = *pixel;
        }

    SDL_UnlockSurface(player_figures_reverse[k]);            
    SDL_UnlockSurface(player_figures[k]);      
  }
  
}

/* vykreslenie konkretnej jednotky */
void drawUnit(int player, int type, SDL_Surface* surface, SDL_Rect rect, int x, int y, bool moving_right) {
  SDL_Rect src = (SDL_Rect){(Sint16)(getUnitTileIndex(type) * FIGURE_WIDTH), 0, FIGURE_WIDTH, FIGURE_HEIGHT};
  SDL_Rect dest = (SDL_Rect){(Sint16)(rect.x + x), (Sint16)(rect.y + y), FIGURE_WIDTH, FIGURE_HEIGHT};
  
  if (moving_right)
    SDL_BlitSurface(player_figures[player], &src, surface, &dest);
  else
    SDL_BlitSurface(player_figures_reverse[player], &src, surface, &dest);
}

/* vykreslenie konkretnej priserky */
void drawMonster(int type, SDL_Surface* surface, SDL_Rect rect, int x, int y, bool moving_right) {
  SDL_Rect src = (SDL_Rect){(Sint16)(type * FIGURE_WIDTH), 0, FIGURE_WIDTH, FIGURE_HEIGHT};
  SDL_Rect dest = (SDL_Rect){(Sint16)(rect.x + x), (Sint16)(rect.y + y), FIGURE_WIDTH, FIGURE_HEIGHT}; 
  
  if (moving_right)
    SDL_BlitSurface(monster_figures, &src, surface, &dest);
  else
    SDL_BlitSurface(monster_figures_reverse, &src, surface, &dest);    
}

/* vykreslenie krvi */
void drawBlood(SDL_Surface* surface, SDL_Rect rect, int x, int y) {
  SDL_Rect src = (SDL_Rect){0, 0, FIGURE_WIDTH, FIGURE_HEIGHT};
  SDL_Rect dest = (SDL_Rect){(Sint16)(rect.x + x), (Sint16)(rect.y + y), FIGURE_WIDTH, FIGURE_HEIGHT};
  
  SDL_BlitSurface(blood_icon, &src, surface, &dest);
}

/* vykreslenie malej konkretnej jednotky */
void drawSmallUnit(int player, int type, SDL_Surface* surface, SDL_Rect rect, int x, int y) {
  SDL_Rect src = (SDL_Rect){(Sint16)(getUnitTileIndex(type) * FIGURE_ICON_WIDTH), 0, FIGURE_ICON_WIDTH, FIGURE_ICON_HEIGHT};
  SDL_Rect dest = (SDL_Rect){
    (Sint16)(rect.x + x - FIGURE_ICON_X_SHIFT),
    (Sint16)(rect.y + y - FIGURE_ICON_Y_SHIFT),
    FIGURE_ICON_WIDTH,
    FIGURE_ICON_HEIGHT
  };
  
  SDL_BlitSurface(player_small_figures[player], &src, surface, &dest);
}

/* vykreslenie malej konkretnej priserky */
void drawSmallMonster(int type, SDL_Surface* surface, SDL_Rect rect, int x, int y) {
  SDL_Rect src = (SDL_Rect){0, 0, FIGURE_ICON_WIDTH, FIGURE_ICON_HEIGHT};
  SDL_Rect dest = (SDL_Rect){
    (Sint16)(rect.x + x - FIGURE_ICON_X_SHIFT),
    (Sint16)(rect.y + y - FIGURE_ICON_Y_SHIFT),
    FIGURE_ICON_WIDTH,
    FIGURE_ICON_HEIGHT
  };
  
  SDL_BlitSurface(monster_small_figures, &src, surface, &dest);
}

/* vykreslenie maleho predmetu */
void drawSmallItem(int typ, SDL_Surface* surface, SDL_Rect rect, int x, int y) {
  SDL_Rect src = (SDL_Rect){0, 0, (Uint16)(item_small_icon -> w), (Uint16)(item_small_icon -> h)};
  SDL_Rect dest = (SDL_Rect){
    (Sint16)(rect.x + x * TILE_WIDTH + (TILE_WIDTH - item_small_icon -> w) / 2),
    (Sint16)(rect.y + y * TILE_HEIGHT + (TILE_HEIGHT - item_small_icon -> h) / 2),
    (Uint16)(item_small_icon -> w),
    (Uint16)(item_small_icon -> h)
  };
  
  SDL_BlitSurface(item_small_icon, &src, surface, &dest);  
}

/* vykreslenie neseneho predmetu */
void drawCarriedItem(int typ, SDL_Surface* surface, SDL_Rect rect, int x, int y) {  
  SDL_Rect src = (SDL_Rect){0, 0, (Uint16)(item_carried_icon -> w), (Uint16)(item_carried_icon -> h)};
  SDL_Rect dest = (SDL_Rect){
    (Sint16)(rect.x + x + (ZOOM_TILE_WIDTH - item_carried_icon -> w) / 2),
    (Sint16)(rect.y + y + (ZOOM_TILE_HEIGHT - item_carried_icon -> h) / 2),
    (Uint16)(item_carried_icon -> w),
    (Uint16)(item_carried_icon -> h)
  };
  
  SDL_BlitSurface(item_carried_icon, &src, surface, &dest);  
}

/* vykreslenie predmetu na zemi */
void drawFallenItem(int typ, SDL_Surface* surface, SDL_Rect rect, int x, int y) {
  SDL_Rect src = (SDL_Rect){0, 0, (Uint16)(item_big_icon -> w), (Uint16)(item_big_icon -> h)};
  SDL_Rect dest = (SDL_Rect){
    (Sint16)(rect.x + x * ZOOM_TILE_WIDTH + ((int)ZOOM_TILE_WIDTH - item_big_icon -> w) / 2),
    (Sint16)(rect.y + y * ZOOM_TILE_HEIGHT + ((int)ZOOM_TILE_HEIGHT - item_big_icon -> h) / 2),
    (Uint16)(item_big_icon -> w),
    (Uint16)(item_big_icon -> h)
  };
  
  SDL_BlitSurface(item_big_icon, &src, surface, &dest);  
}

/* vykreslenie vsetkych jednotiek a priserok */
void renderFigures(SDL_Surface* surface, SDL_Rect rect, int startX, int startY) {
  const std::vector<Zivocich>& figures = states[game_time / GAMETIME_UNITS].zivocichy;
  const std::vector<Zivocich>& figures_next = states[(game_time / GAMETIME_UNITS) + 1].zivocichy;
  
  const std::vector<Predmet>& items = states[game_time / GAMETIME_UNITS].predmety;
  //const std::vector<Predmet>& items_next = states[(game_time / GAMETIME_UNITS) + 1].predmety;
  
  std::vector<std::pair<int, int> > hits;
  
  for (int i = 0; i < (int)answers[game_time / GAMETIME_UNITS].size(); i++) {
    const Prikaz& prikaz = answers[game_time / GAMETIME_UNITS][i];
    
    if (prikaz.typ != PRIKAZ_UTOC) continue;
    
    hits.push_back(std::make_pair(prikaz.cielX, prikaz.cielY));
  }
      
  double step = (double)(game_time % GAMETIME_UNITS) / GAMETIME_UNITS;
  
  for (int i = 0; i < (int)figures.size(); i++) {
    int x = ZOOM_TILE_WIDTH * (figures[i].x - startX);
    int y = ZOOM_TILE_HEIGHT * (figures[i].y - startY);
            
    bool found = false;
    bool blooded = false;
    
    for (int j = 0; j < (int)hits.size(); j++)
      if (hits[j] == std::make_pair(figures[i].x, figures[i].y))
        blooded = true;
        
    for (int j = 0; j < (int)figures_next.size(); j++)
      if (figures[i].id == figures_next[j].id) {
        x += (int)((figures_next[j].x - figures[i].x) * ZOOM_TILE_WIDTH * step);
        y += (int)((figures_next[j].y - figures[i].y) * ZOOM_TILE_HEIGHT * step);
        
        if (figures_next[j].x > figures[i].x) who_moved_right.insert(figures[i].id);
        if (figures_next[j].x < figures[i].x) who_moved_right.erase(figures[i].id);
        
        found = true;        
      }
      
    if (x <= -ZOOM_TILE_WIDTH || x >= ZOOMMAP_WIDTH * ZOOM_TILE_WIDTH || y <= -ZOOM_TILE_HEIGHT || y >= ZOOMMAP_HEIGHT * ZOOM_TILE_HEIGHT) continue;                  
    if (game_time % GAMETIME_UNITS != 0 && !found) {
      // ak niekto neprezil tah, uz ho nevykresluj... ale krv este hej, obviously
      if (blooded)
        drawBlood(surface, rect, x, y);      
    }
    
    bool moving_right = (who_moved_right.find(figures[i].id) != who_moved_right.end());    
        
    if (figures[i].ktorehoHraca == -1) {
      drawMonster(figures[i].id % NUM_MONSTERS, surface, rect, x, y, moving_right);
    } else {
      drawUnit(figures[i].ktorehoHraca, figures[i].typ % NUM_UNITS, surface, rect, x, y, moving_right);
      
      if (figures[i].nesenyPredmet != -1) {
        drawCarriedItem(figures[i].nesenyPredmet, surface, rect, x, y);
      }
    }
    
    if (blooded)
      drawBlood(surface, rect, x, y);
  }
      
  for (int i = 0; i < (int)items.size(); i++) {
    if (items[i].x < startX || items[i].x >= startX + ZOOMMAP_WIDTH || items[i].y < startY || items[i].y >= startY + ZOOMMAP_HEIGHT)
      continue;    
    
    drawFallenItem(items[i].typ, surface, rect, items[i].x - startX, items[i].y - startY);
  }
}

/* vykresli male jednotky a priserky */
void renderSmallFigures(SDL_Surface* surface, SDL_Rect rect) {
  const std::vector<Zivocich>& figures = states[game_time / GAMETIME_UNITS].zivocichy;
  const std::vector<Zivocich>& figures_next = states[(game_time / GAMETIME_UNITS) + 1].zivocichy;
  
  const std::vector<Predmet>& items = states[game_time / GAMETIME_UNITS].predmety;
  //const std::vector<Predmet>& items_next = states[(game_time / GAMETIME_UNITS) + 1].predmety;
  
  double step = (double)(game_time % GAMETIME_UNITS) / GAMETIME_UNITS;
    
  for (int i = 0; i < (int)figures.size(); i++) {
    int x = TILE_WIDTH * figures[i].x;
    int y = TILE_HEIGHT * figures[i].y;
        
    bool found = false;
    for (int j = 0; j < (int)figures_next.size(); j++)
      if (figures[i].id == figures_next[j].id) {
        x += (int)((figures_next[j].x - figures[i].x) * TILE_WIDTH * step);
        y += (int)((figures_next[j].y - figures[i].y) * TILE_HEIGHT * step);
        
        found = true;
      }
      
    if (game_time % GAMETIME_UNITS != 0 && !found) continue; // ak niekto neprezil tah, uz ho nevykresluj    
    
    if (figures[i].ktorehoHraca == -1) {
      drawSmallMonster(figures[i].id % NUM_MONSTERS, surface, rect, x, y);
    } else {
      drawSmallUnit(figures[i].ktorehoHraca, figures[i].typ % NUM_UNITS, surface, rect, x, y);
    }
  }
    
  for (int i = 0; i < (int)items.size(); i++) {
    drawSmallItem(items[i].typ, surface, rect, items[i].x, items[i].y);
  }
}
