#include "marshal.h"
#include "common.h"
#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_image.h"
#include "observer.h"
#include "mapa.h"
#include <cstdio>
#include <cstdlib>
#include <vector>
#include <fstream>
using namespace std;

Mapa world;
SDL_Surface* map_background;
SDL_Surface* zoommap_background;

int MAP_WIDTH = 0;
int MAP_HEIGHT = 0;

/* pozicia textury v tiles.png */
int textureIndex(int type) {
  switch (type) {
    case MAPA_TEMNE_HORY: return 0;    
    case MAPA_HORY: return 1;
    case MAPA_VODA: return 2;
    case MAPA_TRAVA: return 3;
    case MAPA_TEMNA_ZEM: return 4;    
  }  
  
  return 0;
}

void loadMap() {
  /* nacitanie mapy */
  ifstream map_file (getLogFile("map"), ifstream::in);
  nacitaj(map_file, world);
    
  NUM_PLAYERS = world.pocetHracov;
  MAP_WIDTH = world.w;
  MAP_HEIGHT = world.h;
      
  /* otexturovanie malej mapy */
  SDL_Surface* textures = IMG_Load(getFile("tiles/tiles.png"));
  SDL_LockSurface(map_background);
  SDL_LockSurface(textures);
       
  for (int i = 0; i < MAP_WIDTH; i++)
    for (int j = 0; j < MAP_HEIGHT; j++) {
      
      int tIndex = textureIndex(world.pole[j][i]);
      
      for (int m = 0; m < TILE_WIDTH; m++)
        for (int n = 0; n < TILE_HEIGHT; n++) {
          int xoffset = i * TILE_WIDTH + m;
          int yoffset = j * TILE_HEIGHT + n;
                    
          int txoffset = (xoffset % TEXTURE_WIDTH) + (tIndex * TEXTURE_WIDTH);
          int tyoffset = yoffset % TEXTURE_HEIGHT;
          
          Uint32* dest = ((Uint32*)(map_background -> pixels)) + (yoffset * map_background -> w) + xoffset;
          Uint32* src = ((Uint32*)(textures -> pixels)) + (tyoffset * textures -> w) + txoffset;
          *dest = *src;
          
          /* quick fix travy */
          if (world.pole[j][i] == MAPA_TRAVA)
            *(SDL_Color*)dest = (SDL_Color){0x60, 0xC5, 0x80, 0xFF};
        }
    }
      
  SDL_UnlockSurface(textures);
  SDL_UnlockSurface(map_background);
  
  /* pripravenie zoom mapy */
  zoommap_background = SDL_CreateRGBSurface(SDL_SWSURFACE, MAP_WIDTH * ZOOM_TILE_WIDTH, MAP_HEIGHT * ZOOM_TILE_HEIGHT, 32, rmask, gmask, bmask, 0);
  
  for (int i = 0; i < MAP_WIDTH; i++)
    for (int j = 0; j < MAP_HEIGHT; j++) {
      SDL_Rect dest_rect = (SDL_Rect){               
        (Sint16)(i * ZOOM_TILE_WIDTH),
        (Sint16)(j * ZOOM_TILE_HEIGHT),
        ZOOM_TILE_WIDTH,
        ZOOM_TILE_HEIGHT
      };
      
      SDL_Rect src_rect = (SDL_Rect) {
        (Sint16)(textureIndex(world.pole[j][i]) * TEXTURE_WIDTH),
        0,
        TEXTURE_WIDTH,
        TEXTURE_HEIGHT
      };
      
      SDL_BlitSurface(textures, &src_rect, zoommap_background, &dest_rect);
    }
  
  SDL_FreeSurface(textures);
    
  /* ikonky mesta na zoom mape */
  SDL_Surface* city_img = IMG_Load(getFile("tiles/mesto.png"));
  SDL_Surface* small_city_img = IMG_Load(getFile("tiles/mesto_small.png"));

  for (int i = 0; i < MAP_WIDTH; i++)
    for (int j = 0; j < MAP_HEIGHT; j++)
      if (world.pole[j][i] == MAPA_MESTO) {
        SDL_Rect rect = (SDL_Rect){
          (Sint16)(i * ZOOM_TILE_WIDTH - ZOOM_TILE_WIDTH),
          (Sint16)(j * ZOOM_TILE_HEIGHT),
          (Uint16)ZOOM_TILE_WIDTH * 3,
          (Uint16)ZOOM_TILE_HEIGHT
        };
        
        SDL_BlitSurface(city_img, NULL, zoommap_background, &rect);

        rect = (SDL_Rect){
          (Sint16)(i * TILE_WIDTH - (small_city_img -> w / 2)),
          (Sint16)(j * TILE_HEIGHT - (small_city_img -> h / 2)),
          (Uint16)(small_city_img -> w),
          (Uint16)(small_city_img -> h)
        };        
        
        SDL_BlitSurface(small_city_img, NULL, map_background, &rect);
      }
      
  SDL_FreeSurface(city_img);
  SDL_FreeSurface(small_city_img);
}

void drawZoomMap(SDL_Surface* surface, SDL_Rect dest_rect, SDL_Surface* bigmap_surface, SDL_Rect bigmap_rect) {
  /* vyrataj zobrazovanu oblast */
  int startX = interestingX - (ZOOMMAP_WIDTH / 2);
  int startY = interestingY - (ZOOMMAP_HEIGHT / 2);
  
  if (startX + ZOOMMAP_WIDTH >= MAP_WIDTH)
    startX = MAP_WIDTH - ZOOMMAP_WIDTH;
  if (startY + ZOOMMAP_HEIGHT >= MAP_HEIGHT)
    startY = MAP_HEIGHT - ZOOMMAP_HEIGHT;
  
  if (startX < 0) startX = 0;
  if (startY < 0) startY = 0;
  
  /* vykresli mapu */
  SDL_Rect src_rect = (SDL_Rect) {
    Sint16(startX * ZOOM_TILE_WIDTH),
    Sint16(startY * ZOOM_TILE_HEIGHT),
    ZOOMMAP_WIDTH * ZOOM_TILE_WIDTH,
    ZOOMMAP_HEIGHT * ZOOM_TILE_HEIGHT
  };
  
  SDL_BlitSurface(zoommap_background, &src_rect, surface, &dest_rect);
  
  /* vykresli figurky */
  renderFigures(surface, dest_rect, startX, startY);
  
  /* vyznac plochu na velkej mape*/
  SDL_Rect aux_rect = (SDL_Rect){
    (Sint16)(startX * TILE_WIDTH + bigmap_rect.x - MARKRECT_WIDTH),
    (Sint16)(startY * TILE_WIDTH + bigmap_rect.y - MARKRECT_WIDTH),
    ZOOMMAP_WIDTH * TILE_WIDTH + 2 * MARKRECT_WIDTH,
    MARKRECT_WIDTH
  };
  
  SDL_FillRect(bigmap_surface, &aux_rect, MARKRECT_COLOR);
  aux_rect.w = MARKRECT_WIDTH;
  aux_rect.h = ZOOMMAP_HEIGHT * TILE_HEIGHT + 3 * MARKRECT_WIDTH; 
  SDL_FillRect(bigmap_surface, &aux_rect, MARKRECT_COLOR);  
  aux_rect.x = (Sint16)((startX + ZOOMMAP_WIDTH) * TILE_WIDTH + bigmap_rect.x + MARKRECT_WIDTH);
  SDL_FillRect(bigmap_surface, &aux_rect, MARKRECT_COLOR);
  aux_rect.x = (Sint16)(startX * TILE_WIDTH + bigmap_rect.x - MARKRECT_WIDTH);
  aux_rect.y = (Sint16)((startY + ZOOMMAP_HEIGHT) * TILE_HEIGHT + bigmap_rect.y + MARKRECT_WIDTH);
  aux_rect.w = ZOOMMAP_WIDTH * TILE_WIDTH + 2 * MARKRECT_WIDTH;
  aux_rect.h = MARKRECT_WIDTH;
  SDL_FillRect(bigmap_surface, &aux_rect, MARKRECT_COLOR);
}
