#include "marshal.h"
#include "common.h"
#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_image.h"
#include "observer.h"
#include <unistd.h>
#include <time.h>
#include <string>

SDL_Surface* screen;     // toto sa vykresluje

SDL_Surface* rightbar;  // tu sa vykresluju questy
SDL_Surface* infobar;   // tu sa vykresluje dalsie info

SDL_Surface* wooden_background;
SDL_Surface* zoommap_frame;

SDL_Rect aux_rect;
SDL_Rect basicmap_rect = (SDL_Rect){0, 0, DISPLAY_WIDTH - RIGHTBAR_WIDTH, DISPLAY_HEIGHT - INFOBAR_HEIGHT};
SDL_Rect rightbar_rect = (SDL_Rect){DISPLAY_WIDTH - RIGHTBAR_WIDTH, 0, RIGHTBAR_WIDTH, DISPLAY_HEIGHT};
SDL_Rect infobar_rect = (SDL_Rect){0, DISPLAY_HEIGHT - INFOBAR_HEIGHT, DISPLAY_WIDTH - RIGHTBAR_WIDTH, INFOBAR_HEIGHT};
SDL_Rect zoommap_rect = (SDL_Rect){10, 10, ZOOMMAP_WIDTH * ZOOM_TILE_WIDTH, ZOOMMAP_HEIGHT * ZOOM_TILE_HEIGHT};
SDL_Rect infotable_rect = (SDL_Rect){ ZOOMMAP_WIDTH * ZOOM_TILE_WIDTH + 50, 10, 0, 0 };

TTF_Font *bigfont, *smallfont, *hugefont;

int game_time = 0;                                        // GAMETIME_UNITS jednotiek = 1 tah
int game_time_increment = GAMETIME_DEFAULT_INCREMENT;     // d game_time / d frame
bool game_paused = false;

void draw_scene() {
  /* vykreslenie questov, ak treba */
  if (game_time % GAMETIME_UNITS < game_time_increment || game_time % GAMETIME_UNITS > GAMETIME_UNITS - game_time_increment) {
    SDL_BlitSurface(wooden_background, NULL, rightbar, NULL);  
    drawQuests(rightbar, states[game_time / GAMETIME_UNITS].questy);
  }
  
  /* vykreslenie sceny v case game_time */
  SDL_BlitSurface(wooden_background, &infobar_rect, infobar, NULL);
  SDL_BlitSurface(zoommap_frame, NULL, infobar, NULL);

  SDL_BlitSurface(map_background, NULL, screen, &basicmap_rect);  
  renderSmallFigures(screen, basicmap_rect);  
  
  drawZoomMap(infobar, zoommap_rect, screen, basicmap_rect);
  renderAdditionalInfo(infobar, infotable_rect);
  
  SDL_BlitSurface(infobar, NULL, screen, &infobar_rect);
  SDL_BlitSurface(rightbar, NULL, screen, &rightbar_rect);
      
  SDL_Flip(screen);
}

int main(int argc, char* argv[]) {
  /* nacitaj data */
  if (argc < 2) {
    fprintf(stderr, "Usage: observer [log number]\n");
    return 1;
  }  
  loadData(argv[0], argv[1]);
  
  /* inicializacia nahodnosti */
  srand(time(NULL));
  
  /* inicializacia grafiky */
  SDL_Init(SDL_INIT_VIDEO);
  atexit(SDL_Quit);
  
  screen = SDL_SetVideoMode(DISPLAY_WIDTH, DISPLAY_HEIGHT, 32, SDL_SWSURFACE | SDL_DOUBLEBUF /*| SDL_NOFRAME | SDL_FULLSCREEN */);
  
  std::string title_string = "Proboj KSP Trusalova 2013       ";
  title_string += argv[1];
  SDL_WM_SetCaption(title_string.c_str(), 0);
  
  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  /* nacitanie obrazkov */
  wooden_background = IMG_Load(getFile("backgrounds/wood.jpg"));
  zoommap_frame = IMG_Load(getFile("backgrounds/zoommap_frame.png"));
  loadQuestImages();

  /* pripravenie platien */
  rightbar = SDL_CreateRGBSurface(SDL_SWSURFACE, rightbar_rect.w, rightbar_rect.h, 32, rmask, gmask, bmask, 0);
  infobar = SDL_CreateRGBSurface(SDL_SWSURFACE, infobar_rect.w, infobar_rect.h, 32, rmask, gmask, bmask, 0);
  map_background = SDL_CreateRGBSurface(SDL_SWSURFACE, basicmap_rect.w, basicmap_rect.h, 32, rmask, gmask, bmask, 0);
    
  /* nacitanie vykreslovacich modulov */
  loadMap();
  loadFigures();
  
  /* inicializacia fontov */
  TTF_Init();
  atexit(TTF_Quit);
  bigfont = TTF_OpenFont(getFile("fonts/FreeSansBold.ttf"), 24);
  smallfont = TTF_OpenFont(getFile("fonts/FreeSansBold.ttf"), 16);
  hugefont = TTF_OpenFont(getFile("fonts/FreeSansBold.ttf"), 72);
    
  /* hlavny cyklus */
  while (true) {
    /* najdi zaujimavosti, vykresli a pockaj */
    trackInterest();
    draw_scene();
    compensateFPS();
    
    /* posun scenu */
    if (!game_paused)
      game_time += game_time_increment;

    /* zastav spatne premietanie na prvom tahu */    
    if (game_time < 0) {
      game_time = 0;
      game_time_increment = 0;
    }
    
    /* ukonci hru pri poslednom tahu */
    if (game_time >= TURNS_COUNT * GAMETIME_UNITS) {
      game_time = TURNS_COUNT * GAMETIME_UNITS;
      draw_scene();
      
      sleep(ENDGAME_DELAY);
      break;
    }
    
    /* reakcia na klavesy */
    SDL_Event event;

    while (SDL_PollEvent(&event)) {
      if (event.type == SDL_KEYDOWN)
        switch(event.key.keysym.sym) {
          case SDLK_c: // nasilne vypnutie
            if (event.key.keysym.mod & KMOD_CTRL)
              return -1;
          break;
          
          case SDLK_s: // normalne vypnutie
            if (event.key.keysym.mod & KMOD_CTRL)
              return 0;
          break;
          
          case SDLK_p: // pauza
            game_paused = !game_paused;
          break;
          
          case SDLK_KP_PLUS:
          case SDLK_PLUS: // zvys rychlost premietania
            game_time_increment += GAMETIME_CHANGERATE;
          break;
          
          case SDLK_KP_MINUS:
          case SDLK_MINUS: // zniz rychlost premietania
            game_time_increment -= GAMETIME_CHANGERATE;
          break;
          
          case SDLK_DOWN:
          case SDLK_UP:
          case SDLK_LEFT:
          case SDLK_RIGHT:
            keyTrack(event.key);
          
          default:
          break;
        }
        
      if (event.type == SDL_MOUSEBUTTONDOWN) {
        mouseTrack(event.button, basicmap_rect);
      }

      if (event.type == SDL_QUIT)
        return -1;      
    }
  }
     
  return 0;
}
