#include "marshal.h"
#include "common.h"
#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_image.h"
#include "observer.h"
#include <cstdlib>

int interestingX;
int interestingY;

static int trackId = -1;
int idling = getMiliseconds();

static void trackNear(int resX, int resY) {
  trackId = -1;
  int bestdist = 12*12;
  const std::vector<Zivocich>& zivocichy = states[game_time / GAMETIME_UNITS].zivocichy;
  FOREACH(it, zivocichy) {
    if (it->ktorehoHraca == -1) continue;
    int d = (it->x-resX)*(it->x-resX) + (it->y-resY)*(it->y-resY);
    if (d < bestdist) {
      bestdist = d;
      trackId = it->id;
      interestingX = it->x;
      interestingY = it->y;
    }
  }
  if (trackId != -1) idling = getMiliseconds();
}

void trackInterest() {
  /* zaujem sa nahodne nastavi na zaciatku a po styroch sekundach idlenia
   * od posledneho nastavenia (ci uz nahodneho, alebo ineho)
   * 
   */

  const std::vector<Zivocich>& figures = states[game_time / GAMETIME_UNITS].zivocichy;

  if (trackId != -1) {
    FOREACH (it, figures) {
      if (it->id == trackId) {
        interestingX = it->x;
        interestingY = it->y;
        idling = getMiliseconds();
        return;
      }
    }
    trackNear(interestingX, interestingY);
    return;
  }
  
  if (game_time == 0 && game_time_increment != 0) {}
  else if (getMiliseconds() - idling < IDLING_MILLISECONDS) return;
  idling = getMiliseconds();
    
  if (figures.size()) {
    int index = rand() % figures.size();  
    interestingX = figures[index].x; 
    interestingY = figures[index].y;
  } else {
    interestingX = rand() % MAP_WIDTH;
    interestingY = rand() % MAP_HEIGHT;
  }
}

void mouseTrack(SDL_MouseButtonEvent event, SDL_Rect map_rect) {
  int resX = (event.x - map_rect.x) / TILE_WIDTH;
  int resY = (event.y - map_rect.y) / TILE_HEIGHT;
  
  if (resX >= MAP_WIDTH) return;
  if (resY >= MAP_HEIGHT) return;
  
  if (event.button == SDL_BUTTON_RIGHT) {
    trackNear(resX, resY);
    return;
  }

  trackId = -1;
  interestingX = resX;
  interestingY = resY;
  
  idling = getMiliseconds();
}

void keyTrack(SDL_KeyboardEvent event) {
  int resX = interestingX;
  int resY = interestingY;
  
  int inc = 3;
  switch(event.keysym.sym) {
    case SDLK_DOWN:
      resY += inc;
    break;
    
    case SDLK_UP:
      resY -= inc;
    break;
    
    case SDLK_LEFT:
      resX -= inc;
    break;
    
    case SDLK_RIGHT:
      resX += inc;
    break;
    
    default:
    break;
  }
  
  int startX = resX - (ZOOMMAP_WIDTH / 2);
  int startY = resY - (ZOOMMAP_HEIGHT / 2);
    
  if (startX + ZOOMMAP_WIDTH >= MAP_WIDTH)
    startX = MAP_WIDTH - ZOOMMAP_WIDTH;
  if (startY + ZOOMMAP_HEIGHT >= MAP_HEIGHT)
    startY = MAP_HEIGHT - ZOOMMAP_HEIGHT;
  
  if (startX < 0) startX = 0;
  if (startY < 0) startY = 0;
  
  interestingX = startX + (ZOOMMAP_WIDTH / 2);
  interestingY = startY + (ZOOMMAP_HEIGHT / 2);
  
  idling = getMiliseconds();
}
