
#include <ostream>
#include <vector>
#include <queue>
#include <set>
#include <algorithm>
#include <cmath>
using namespace std;

#include "common.h"
#include "update.h"

// clockwise zhora
const int DX[] = { 0, 1, 1, 1, 0, -1, -1, -1 };
const int DY[] = { -1, -1, 0, 1, 1, 1, 0, -1 };

const int kApZasoba = 25;
const int kApJedlo = 1;
const int kApChod[HRDINA_POCET_TYPOV][9] = {
  { 2, 2, 3, 3, 4, 4, 5, 5, -1 },
  { 2, 2, 2, 3, 4, 4, 5, 5, -1 },
  { 1, 2, 2, 3, 3, 4, 5, 5, -1 },
  { 2, 3, 3, 4, 4, 5, 5, 5, -1 },
  { 2, 3, 3, 4, 4, 4, 5, 5, -1 },
};
const int kApUtoc[OBRANA_POCET_STUPNOV][6] = {
  { 1, 3, 4, 5,  6, -1 },
  { 1, 4, 5, 7, 10, -1 },
  { 2, 3, 6, 8, 10, -1 },
  { 3, 4, 6, 8, 10, -1 },
  { 4, 6, 8, 9, 10, -1 },
};
const int kApPoloz = 1;
const int kApVerbuj = 17;

const int kHrdinaObrana[HRDINA_POCET_TYPOV][ZBRAN_POCET_TYPOV] = {
  { 3, 4, 1, 1, 2, 2, 2, 3, 2 },
  { 1, 1, 2, 3, 2, 4, 2, 2, 3 },
  { 2, 2, 3, 2, 1, 2, 3, 1, 4 },
  { 3, 1, 2, 4, 3, 1, 2, 2, 2 },
  { 2, 3, 2, 1, 4, 2, 1, 2, 3 },
};
const int kHrdinaHp[HRDINA_POCET_TYPOV] = { 10, 9, 8, 7, 8 };
const int kMaximumHrdinovHraca = 4;
const int kPeriodaVerbovania = 9;

const int kPotvoraDohlad = 4;
const int kPotvoraLambdaChodim = 2;
const int kPotvoraLambdaUtocim = 4;
const int kPotvoraObranaSance[HRDINA_POCET_TYPOV][5] = {
  {  5, 25, 40, 25,  5 },
  {  0, 30, 30, 30, 10 },
  {  0, 10, 20, 40, 30 },
  { 10, 40, 25, 15, 10 },
  { 40,  5,  5, 20, 30 },
};
const int kSpawnFaktor = 100;
const int kSpawnPokusy = 25;

const int kQuestyMaximum = 20;
const int kQuestyPeriodaZaklad = 7;
const int kQuestyPeriodaFaktor = 100;
const int kQuestyVzacne = 65;
const int kQuestyPrchavost = 9;

const int kOdmenaBezna = 15;
const int kOdmenaPrenos = 20;
const int kOdmenaNajdi = 25;
const int kOdmenaNajdiBonus = 15;
const int kOdmenaZnicFaktor = 14;

const int kPredmetyBezne = 20;
const int kPredmetyVzacne = 30;
const int kPredmetyPeriodaZaklad = 20;
const int kPredmetyPeriodaFaktor = 500;

const int kDlzkaHry = 300;


static ostream* g_observation;
void zapniObservation(ostream* observation) { g_observation = observation; }

#define OBSERVE(s,...) do {                                                    \
    if (!g_observation) break;                                                 \
    *g_observation << (s);                                                     \
    int __m[] = { __VA_ARGS__ };                                               \
    for (unsigned __i = 0; __i < sizeof(__m)/sizeof(*__m); __i++)              \
      *g_observation << " " << __m[__i];                                       \
    *g_observation << endl;                                                    \
  } while(0)


Stav zaciatokHry(const Mapa& mapa) {
  Stav stav = Stav();   // tento zapis vsetko inicializuje na nuly
  stav.hraci.resize(mapa.pocetHracov);
  return stav;
}


Tah zaciatokTahu(const Stav& stav, int hrac) {
  Tah tah = Tah();   // tento zapis vsetko inicializuje na nuly
  tah.hrac = hrac;
  tah.ap = kApZasoba;
  FOREACH (it, stav.zivocichy) {
    if (it->ktorehoHraca == hrac) tah.ap -= kApJedlo;
  }
  return tah;
}


int vzdialenost(int x1, int y1, int x2, int y2) {
  return max(abs(x1 - x2), abs(y1 - y2));
}


template<class T> unsigned najdiXy(const vector<T>& v, int x, int y) {
  for (unsigned i = 0; i < v.size(); i++) if (v[i].x == x && v[i].y == y) return i;
  return v.size();
}


static unsigned najdiZivocicha(const vector<Zivocich>& v, int id) {
  for (unsigned i = 0; i < v.size(); i++) if (v[i].id == id) return i;
  return v.size();
}


template<class K,class V> int constMapRead(const map<K,V>& m, const K& key, const V& def) {
  typeof(m.begin()) it = m.find(key);
  if (it == m.end()) return def;
  return it->second;
}


int zistiApCenu(const Mapa& mapa, const Stav& stav, const Tah& tah, const Prikaz& p) {
  if (p.typ == PRIKAZ_VERBUJ) {
    if (tah.hrac == -1) return -1;
    if (p.kto < 0 || p.kto >= HRDINA_POCET_TYPOV) return -1;
    if (mapa.zisti(p.cielX, p.cielY) != MAPA_MESTO) return -1;
    set<int> mojeTypy;
    FOREACH (it, stav.pracovnePonuky) if (it->ktorehoHraca == tah.hrac) mojeTypy.insert(it->typ);
    FOREACH (it, stav.zivocichy) if (it->ktorehoHraca == tah.hrac) mojeTypy.insert(it->typ);
    if (mojeTypy.count(p.kto) || mojeTypy.size() >= (unsigned)kMaximumHrdinovHraca) return -1;
    return kApVerbuj;
  }

  unsigned zivocich = najdiZivocicha(stav.zivocichy, p.kto);
  if (zivocich == stav.zivocichy.size()) return -1;
  const Zivocich& z = stav.zivocichy[zivocich];
  if (z.ktorehoHraca != tah.hrac) return -1;

  int range = 1;   // TODO iny range pri niektorych utokoch?
  if (vzdialenost(z.x, z.y, p.cielX, p.cielY) > range) return -1;
  if (!mapa.priechodne(p.cielX, p.cielY)) return -1;

  bool jeTamZivocich = (najdiXy(stav.zivocichy, p.cielX, p.cielY) != stav.zivocichy.size());
  bool jeTamPredmet = (najdiXy(stav.predmety, p.cielX, p.cielY) != stav.predmety.size());

  if (p.typ == PRIKAZ_POLOZ) {
    if (tah.hrac == -1) return -1;
    if (z.nesenyPredmet == -1) return -1;
    if (jeTamZivocich || jeTamPredmet) return -1;
    return kApPoloz;
  }

  if (p.typ == PRIKAZ_CHOD) {
    if (jeTamZivocich) return -1;
    if (z.nesenyPredmet != -1 && jeTamPredmet) return -1;
    if (tah.hrac == -1 && jeTamPredmet) return -1;
    int typ = tah.hrac == -1 ? HRDINA_BOJOVNIK : z.typ;
    int pocet = constMapRead(tah.ktoKolkoChodil, z.id, 0);
    return kApChod[typ][pocet];
  }

  if (p.typ == PRIKAZ_UTOC) {
    if (!jeTamZivocich) return -1;
    unsigned ciel = najdiXy(stav.zivocichy, p.cielX, p.cielY);
    const Zivocich& c = stav.zivocichy[ciel];
    if (z.ktorehoHraca == -1 && c.ktorehoHraca == -1) return -1;
    int obrana = c.obrana[z.typ];
    int pocet = constMapRead(tah.ktoKolkoUtocil, z.id, 0);
    return kApUtoc[obrana][pocet];
  }

  return -1;
}


bool mozeVykonatPrikaz(const Mapa& mapa, const Stav& stav, const Tah& tah, const Prikaz& p) {
  int cena = zistiApCenu(mapa, stav, tah, p);
  if (cena < 0 || cena > tah.ap) return false;
  return true;
}


void zoberPredmet(const Mapa& mapa, Stav& stav, Zivocich& z) {
  unsigned predmet = najdiXy(stav.predmety, z.x, z.y);
  if (predmet == stav.predmety.size()) return;
  Predmet& p = stav.predmety[predmet];
  OBSERVE("zoberPredmet", z.id, p.x, p.y, p.typ);
  z.nesenyPredmet = p.typ;
  ERASE(stav.predmety, stav.predmety[predmet]);
}


void polozPredmet(const Mapa& mapa, Stav& stav, Zivocich& z, int x, int y) {
  FOREACH (it, stav.questy) {
    if (vzdialenost(x, y, it->cielX, it->cielY) <= 1 && it->ziadanyPredmet == z.nesenyPredmet) {
      stav.hraci[z.ktorehoHraca].zlato += it->zlato;
      OBSERVE("polozPredmet.splnenyQuest", z.id, it->cielX, it->cielY, it->ziadanyPredmet);
      ERASE(stav.questy, *it);
      z.nesenyPredmet = -1;
      return;
    }
  }
  OBSERVE("polozPredmet", z.id, x, y, z.nesenyPredmet);
  Predmet p;
  p.x = x;
  p.y = y;
  p.typ = z.nesenyPredmet;
  stav.predmety.push_back(p);
  z.nesenyPredmet = -1;
}


void zautoc(const Mapa& mapa, Stav& stav, int kto, int ciel) {
  Zivocich& z = stav.zivocichy[kto];
  Zivocich& c = stav.zivocichy[ciel];
  OBSERVE("zautoc", z.id, c.id);
  c.hp--;
  if (c.hp == 0) {
    OBSERVE("zautoc.zabil", z.id, c.id);
    if (c.nesenyPredmet != -1) {
      if (z.nesenyPredmet == -1) {
        OBSERVE("zautoc.zobral", z.id, c.id, c.nesenyPredmet);
        z.nesenyPredmet = c.nesenyPredmet;
        c.nesenyPredmet = -1;
      }
      else {
        polozPredmet(mapa, stav, c, c.x, c.y);
      }
    }
    ERASE(stav.zivocichy, c);
  }
  else if (c.ktorehoHraca == -1) {
    c.potvoraNalada = NALADA_UTOCIM;
  }
}


bool vykonajPrikaz(const Mapa& mapa, Stav& stav, Tah& tah, const Prikaz& p) {
  int cena = zistiApCenu(mapa, stav, tah, p);
  if (cena < 0 || cena > tah.ap) return false;
  tah.ap -= cena;
  tah.prikazy.push_back(p);
  OBSERVE("vykonajPrikaz", p.typ, p.kto, p.cielX, p.cielY);

  switch (p.typ) {
    case PRIKAZ_CHOD: {
      unsigned zivocich = najdiZivocicha(stav.zivocichy, p.kto);
      Zivocich& z = stav.zivocichy[zivocich];
      tah.ktoKolkoChodil[z.id]++;

      OBSERVE("chod", z.id, z.x, z.y, p.cielX, p.cielY);
      z.x = p.cielX;
      z.y = p.cielY;
      zoberPredmet(mapa, stav, z);
    }
    break;

    case PRIKAZ_UTOC: {
      unsigned zivocich = najdiZivocicha(stav.zivocichy, p.kto);
      unsigned ciel = najdiXy(stav.zivocichy, p.cielX, p.cielY);
      Zivocich& z = stav.zivocichy[zivocich];
      tah.ktoKolkoUtocil[z.id]++;

      zautoc(mapa, stav, zivocich, ciel);
    }
    break;

    case PRIKAZ_POLOZ: {
      unsigned zivocich = najdiZivocicha(stav.zivocichy, p.kto);
      Zivocich& z = stav.zivocichy[zivocich];
      polozPredmet(mapa, stav, z, p.cielX, p.cielY);
    }
    break;

    case PRIKAZ_VERBUJ: {
      Zivocich z = Zivocich();
      z.id = (stav.dalsiId++);
      z.x = z.povodX = p.cielX;
      z.y = z.povodY = p.cielY;
      z.ktorehoHraca = tah.hrac;
      z.typ = p.kto;
      z.hp = kHrdinaHp[p.kto];
      z.nesenyPredmet = -1;
      for (int i = 0; i < ZBRAN_POCET_TYPOV; i++) z.obrana.push_back(kHrdinaObrana[z.typ][i]);
      stav.pracovnePonuky.push_back(z);
      OBSERVE("vykonajPrikaz.verbuj", z.id);
    }
    break;
  }

  return true;
}


bool vykonajPrijatePrikazy(const Mapa& mapa, Stav& stav, int hrac, vector<Prikaz>& prikazy) {
  Tah tah = zaciatokTahu(stav, hrac);
  for (unsigned i = 0; i < prikazy.size(); i++) {
    if (!vykonajPrikaz(mapa, stav, tah, prikazy[i])) {
      prikazy.resize(i);
      return false;
    }
  }
  return true;
}


Prikaz coRobiPotvora(const Mapa& mapa, Stav& stav, const Tah& umelyTah, int id) {
  // "return Prikaz::poloz(id, -1, -1)" tu znamena koniec tahu potvory

  unsigned zivocich = najdiZivocicha(stav.zivocichy, id);
  if (zivocich == stav.zivocichy.size()) return Prikaz::poloz(id, -1, -1);
  Zivocich& z = stav.zivocichy[zivocich];

  vector<pair<int,int> > susedim, vidim;
  FOREACH (it, stav.zivocichy) {
    if (it->ktorehoHraca == -1) continue;
    if (vzdialenost(z.x, z.y, it->x, it->y) <= 1) susedim.push_back(make_pair(it->x, it->y));
    else if (vzdialenost(z.x, z.y, it->x, it->y) <= kPotvoraDohlad) vidim.push_back(make_pair(it->x, it->y));
  }

  if (z.potvoraNalada == NALADA_SPIM) return Prikaz::poloz(id, -1, -1);

  if (z.potvoraNalada == NALADA_CHODIM) {
    susedim.clear();
    vidim.clear();
  }

  // susedim s hrdinom -> utocim
  if (susedim.size()) {
    int r = rand() % susedim.size();
    return Prikaz::utoc(id, susedim[r].first, susedim[r].second);
  }

  // vidim hrdinu -> nahanam
  if (vidim.size()) {
    int r = rand() % vidim.size();
    int tx = vidim[r].first, ty = vidim[r].second;
    int nx = z.x, ny = z.y;
    do {
      if (tx < nx && rand() % 2) nx--;
      if (tx > nx && rand() % 2) nx++;
      if (ty < ny && rand() % 2) ny--;
      if (ty > ny && rand() % 2) ny++;
    } while(nx == z.x && ny == z.y);
    return Prikaz::chod(id, nx, ny);
  }

  // nevidim nikoho -> prechadzam sa
  vector<pair<int,int> > moznosti;
  for (int d = 0; d < 8; d++) {
    int nx = z.x + DX[d], ny = z.y + DY[d];
    if (zistiApCenu(mapa, stav, umelyTah, Prikaz::chod(id, nx, ny)) == -1) continue;
    if (najdiXy(stav.predmety, nx, ny) != stav.predmety.size()) continue;
    int faktor = (vzdialenost(nx, ny, z.povodX, z.povodY) > kPotvoraDohlad ? 1 : 3);
    for (int i = 0; i < faktor; i++) moznosti.push_back(make_pair(nx, ny));
  }
  if (moznosti.empty()) return Prikaz::poloz(id, -1, -1);
  int r = rand() % moznosti.size();
  return Prikaz::chod(id, moznosti[r].first, moznosti[r].second);
}


int poissonRandom(double lambda) {
  // Knuthov algoritmus (magia)
  double L = exp(-lambda);
  int k = 0;
  double p = 1;
  do {
    k++;
    p *= (1.0 * rand() / RAND_MAX);
  } while (p > L);
  return k - 1;
}


void tahajPotvorou(const Mapa& mapa, Stav& stav, int id) {
  unsigned zivocich = najdiZivocicha(stav.zivocichy, id);
  if (zivocich == stav.zivocichy.size()) return;
  Zivocich& z = stav.zivocichy[zivocich];

  if (z.potvoraNalada == NALADA_SPIM) {
    FOREACH (it, stav.zivocichy) {
      if (vzdialenost(z.x, z.y, it->x, it->y) <= 1 && it->ktorehoHraca != -1) {
        z.potvoraNalada = NALADA_CHODIM;
      }
    }
    return;
  }

  Tah tah = zaciatokTahu(stav, -1);
  tah.ap = poissonRandom(z.potvoraNalada == NALADA_UTOCIM ? kPotvoraLambdaUtocim : kPotvoraLambdaChodim);

  while (true) {
    Prikaz p = coRobiPotvora(mapa, stav, tah, id);
    if (p.typ == PRIKAZ_POLOZ) break;   // potvory tymto znacia koniec tahu
    if (!vykonajPrikaz(mapa, stav, tah, p)) break;
  }
}


int spawnujPotvoru(const Mapa& mapa, Stav& stav, int hrac, int vzacnyPredmet) {
  int gx = rand() % mapa.w;
  int gy = rand() % mapa.h;
  if (mapa.zisti(gx, gy) != MAPA_TEMNA_ZEM && rand() % 100 < 60) return 0;

  int typ = rand() % ZBRAN_POCET_TYPOV;
  int pocet = (rand() % 100 < 20) ? (2 + rand() % 5) : 1;
  int hpUroven = rand() % 3;
  int nalada = rand() % 3;
  int beznyPredmet = rand() % kPredmetyBezne;
  if (pocet > 1) {
    hpUroven = rand() % 2;
    nalada = (rand() % 100 < 30 ? NALADA_CHODIM : NALADA_UTOCIM);
    if (rand() % 100 < 30) beznyPredmet = -1;
  }

  vector<int> obrana;
  for (int i = 0; i < HRDINA_POCET_TYPOV; i++) {
    int hodnota = rand() % 100;
    for (int j = 0; j < 5; j++) {
      if (hodnota < kPotvoraObranaSance[i][j]) {
        obrana.push_back(j);
        break;
      }
      hodnota -= kPotvoraObranaSance[i][j];
    }
  }

  int uspechov = 0;
  for (int i = 0; i < pocet; i++) {
    int x = gx + rand() % 5 - 2;
    int y = gy + rand() % 5 - 2;
    int hp = 4 * hpUroven + 1 + rand() % 4;
    if (pocet > 1) hp /= 2;
    if (!mapa.priechodne(x, y)) continue;
    bool zle = false;
    int pobliz = 0;
    if (najdiXy(stav.predmety, x, y) != stav.predmety.size()) zle = true;
    FOREACH (it, stav.zivocichy) {
      if (vzdialenost(x, y, it->x, it->y) <= 1) zle = true;
      if (vzdialenost(x, y, it->x, it->y) <= 5) pobliz++;
      if (it->ktorehoHraca != -1 && vzdialenost(x, y, it->x, it->y) <= 10) zle = true;
    }
    for (int yy = y - 8; yy <= y + 8; yy++) {
      for (int xx = x - 8; xx <= x + 8; xx++) {
        if (mapa.zisti(xx, yy) == MAPA_MESTO) zle = true;
      }
    }
    if (zle) continue;
    if (pobliz > 10) continue;

    int mojPredmet =
      (vzacnyPredmet != -1 && uspechov == 0) ? vzacnyPredmet :
      (rand() % 100 < (pocet > 1 ? 40 : 60)) ? beznyPredmet : -1;

    Zivocich z = Zivocich();
    z.id = (stav.dalsiId++);
    z.x = z.povodX = x;
    z.y = z.povodY = y;
    z.ktorehoHraca = -1;
    z.typ = typ;
    z.nesenyPredmet = mojPredmet;
    z.hp = hp;
    z.obrana = obrana;
    z.potvoraFaza = hrac;
    z.potvoraNalada = nalada;
    stav.zivocichy.push_back(z);
    OBSERVE("spawnujPotvoru", z.id, z.x, z.y, z.typ, z.nesenyPredmet);
    uspechov++;
  }

  if (uspechov == 0) return 0;
  return (nalada == NALADA_UTOCIM ? uspechov : max(uspechov, 2)) * hpUroven * hpUroven;
}


int generujPredmet(const Mapa& mapa, Stav& stav, int predmet) {
  for (int i = 0; i < kSpawnPokusy; i++) {
    int x = rand() % mapa.w;
    int y = rand() % mapa.h;
    bool zle = false;
    if (!mapa.priechodne(x, y)) zle = true;
    if (najdiXy(stav.predmety, x, y) != stav.predmety.size()) zle = true;
    FOREACH (it, stav.zivocichy) {
      if (x == it->x && y == it->y) zle = true;
      if (it->ktorehoHraca != -1 && vzdialenost(x, y, it->x, it->y) <= 3) zle = true;
    }
    for (int yy = y - 8; yy <= y + 8; yy++) {
      for (int xx = x - 8; xx <= x + 8; xx++) {
        if (mapa.zisti(xx, yy) == MAPA_MESTO) zle = true;
      }
    }
    if (zle) continue;

    Predmet p;
    p.x = x;
    p.y = y;
    p.typ = predmet;
    stav.predmety.push_back(p);
    OBSERVE("generujPredmet", x, y, predmet);
    return mapa.zisti(x, y) == MAPA_TEMNA_ZEM ? 2 : 1;
  }

  return 0;
}


bool generujQuest(const Mapa& mapa, Stav& stav, int hrac) {
  vector<pair<int,int> > mesta;
  for (int y = 0; y < mapa.h; y++) for (int x = 0; x < mapa.w; x++) {
    if (mapa.zisti(x, y) == MAPA_MESTO) mesta.push_back(make_pair(x, y));
  }
  if (mesta.empty()) return false;
  int ciel = rand() % mesta.size();

  bool vzacny = (rand() % 100 < kQuestyVzacne);
  int ziadanyPredmet = vzacny ? (kPredmetyBezne + rand() % kPredmetyVzacne) : (rand() % kPredmetyBezne);

  Quest q;
  q.cielX = mesta[ciel].first;
  q.cielY = mesta[ciel].second;
  q.ziadanyPredmet = ziadanyPredmet;
  q.kedyVznikol = stav.cas;
  q.zlato = kOdmenaBezna;

  FOREACH (it, stav.questy) {
    if (it->cielX == q.cielX && it->cielY == q.cielY && it->ziadanyPredmet == q.ziadanyPredmet) {
      return false;
    }
  }

  if (vzacny) {
    int typ = rand() % 4;
    if (typ <= 1) {
      // zabi potvoru
      int narocnost = 0;
      for (int i = 0; i < kSpawnPokusy && !narocnost; i++) {
        narocnost = spawnujPotvoru(mapa, stav, hrac, ziadanyPredmet);
      }
      if (!narocnost) return false;
      q.zlato = narocnost * kOdmenaZnicFaktor;
    }
    else if (typ == 2) {
      // najdi poklad
      int narocnost = generujPredmet(mapa, stav, ziadanyPredmet);
      if (!narocnost) return false;
      q.zlato = kOdmenaNajdi + (narocnost > 1) * kOdmenaNajdiBonus;
    }
    else {
      // prenes predmet
      ERASE(mesta, mesta[ciel]);
      if (mesta.empty()) return false;
      int start = rand() % mesta.size();

      vector<pair<int,int> > moznosti;
      int sx = mesta[start].first, sy = mesta[start].second;
      for (int y = sy - 1; y <= sy + 1; y++) {
        for (int x = sx - 1; x <= sx + 1; x++) {
          if (!mapa.priechodne(x, y)) continue;
          if (najdiXy(stav.predmety, x, y) != stav.predmety.size()) continue;
          if (najdiXy(stav.zivocichy, x, y) != stav.zivocichy.size()) continue;
          moznosti.push_back(make_pair(x, y));
        }
      }
      if (moznosti.empty()) return false;
      int r = rand() % moznosti.size();

      Predmet p;
      p.x = moznosti[r].first;
      p.y = moznosti[r].second;
      p.typ = typ;
      stav.predmety.push_back(p);
      OBSERVE("generujPredmet", p.x, p.y, p.typ);

      q.zlato = kOdmenaPrenos;
    }
  }

  stav.questy.push_back(q);
  OBSERVE("generujQuest", q.cielX, q.cielY, q.ziadanyPredmet, q.zlato);
  return true;
}


void vykonajPozadie(const Mapa& mapa, Stav& stav, int hrac) {
  OBSERVE("vykonajPozadie", stav.cas, hrac);

  // verbovanie
  FOREACH (it, stav.pracovnePonuky) {
    if (rand() % kPeriodaVerbovania) continue;

    bool jeTamZivocich = (najdiXy(stav.zivocichy, it->x, it->y) != stav.zivocichy.size());
    if (jeTamZivocich) {
      vector<int> moznosti;
      for (int d = 0; d < 8; d++) {
        if (najdiXy(stav.zivocichy, it->x + DX[d], it->y + DY[d]) == stav.zivocichy.size()) {
          moznosti.push_back(d);
        }
      }
      if (!moznosti.size()) continue;
      int r = rand() % moznosti.size();
      it->x += DX[moznosti[r]];
      it->y += DY[moznosti[r]];
    }

    stav.zivocichy.push_back(*it);
    Zivocich& z = stav.zivocichy.back();
    OBSERVE("vykonajPozadie.naverbovany", z.id);
    zoberPredmet(mapa, stav, z);

    ERASE(stav.pracovnePonuky, *it);
    --it;
  }

  // vyprchavanie questov
  FOREACH (it, stav.questy) {
    if (rand() % (100 * mapa.pocetHracov) < kQuestyPrchavost) {
      OBSERVE("vykonajPozadie.questZlacnel", it->cielX, it->cielY, it->ziadanyPredmet);
      it->zlato--;
    }
    if (it->zlato <= 0) {
      OBSERVE("vykonajPozadie.questSkoncil", it->cielX, it->cielY, it->ziadanyPredmet);
      ERASE(stav.questy, *it);
      --it;
    }
  }

  // pohyb potvor
  vector<int> tahajucePotvory;
  FOREACH (it, stav.zivocichy) {
    if (it->ktorehoHraca != -1) continue;
    if (it->potvoraFaza != hrac) continue;
    tahajucePotvory.push_back(it->id);
  }
  random_shuffle(tahajucePotvory.begin(), tahajucePotvory.end());
  FOREACH (it, tahajucePotvory) {
    tahajPotvorou(mapa, stav, *it);
  }

  // spawnovanie potvor
  int pocetPotvor = 0;
  FOREACH (it, stav.zivocichy) if (it->ktorehoHraca == -1) pocetPotvor++;
  if (rand() % ((pocetPotvor + 1) * kSpawnFaktor / 100) == 0) {
    for (int i = 0; i < kSpawnPokusy; i++) {
      if (spawnujPotvoru(mapa, stav, hrac, -1)) break;
    }
  }

  // generovanie predmetov
  if (rand() % (kPredmetyPeriodaZaklad + stav.predmety.size() * kPredmetyPeriodaFaktor / 100) == 0) {
    generujPredmet(mapa, stav, rand() % kPredmetyBezne);
  }

  // generovanie questov
  if (stav.questy.size() < (unsigned)kQuestyMaximum) {
    if (rand() % (kQuestyPeriodaZaklad + stav.questy.size() * kQuestyPeriodaFaktor / 100) == 0) {
      generujQuest(mapa, stav, hrac);
    }
  }

  // liecenie
  FOREACH (it, stav.zivocichy) {
    if (it->ktorehoHraca != hrac) continue;
    if (it->hp == kHrdinaHp[it->typ]) continue;
    bool priMeste = (mapa.zisti(it->x, it->y) == MAPA_MESTO);
    for (int d = 0; d < 8; d++) {
      if (mapa.zisti(it->x + DX[d], it->y + DY[d]) == MAPA_MESTO) priMeste = true;
    }
    if (!priMeste) continue;
    int jeTuNepriatel = false;
    FOREACH (it2, stav.zivocichy) {
      if (it2->ktorehoHraca == hrac) continue;
      if (vzdialenost(it->x, it->y, it2->x, it2->y) > 1) continue;
      jeTuNepriatel = true;
    }
    if (jeTuNepriatel) continue;
    OBSERVE("vykonajPozadie.liecenie", it->id);
    it->hp++;
  }

  OBSERVE("vykonajPozadie.koniec", stav.cas, hrac);
  if (hrac + 1 == mapa.pocetHracov) stav.cas++;
}


bool hraSkoncila(const Mapa& mapa, const Stav& stav) {
  return stav.cas >= kDlzkaHry;
}


int zistiSkore(const Stav& stav, int hrac) {
  // TODO bonusy podla ranku
  return stav.hraci[hrac].zlato;
}

